"""
Created on Oct 25, 2016


@author: jurgen
"""

import datetime
import os
import re
import multiprocessing
import logging
import pprint
import traceback
from typing import List, Optional

from rinex_parser import constants as cc

from rinex_parser.ext.convertdate.convertdate import year_doy
from rinex_parser.logger import logger
from rinex_parser.obs_header import Rinex2ObsHeader, Rinex3ObsHeader, RinexObsHeader
from rinex_parser.obs_epoch import RinexEpoch

# from celery.utils.log import get_task_logger


# celery_logger = get_task_logger(__name__)
# celery_logger.setLevel(logging.DEBUG)
celery_logger = logger

__updated__ = "2016-11-16"


class RinexObsReader(object):
    """
    Doc of Class RinexObsReader

    Args:
        datadict: {
            "epochs": [
                {
                    "id": "YYYY-mm-ddTHH:MM:SSZ",
                    "satellites": [
                        {
                            "id": "<Satellite Number>",
                            "observations": {
                                "<Observation Descriptor>": {
                                    "value": ..,
                                    "lli": ..,
                                    "ss": ..
                                }
                            }
                        },
                        {
                            "id": "...",
                            "observations": {...}
                        }
                    ]
                },
                {
                    "id": "..."
                    "satellites": [..]
                },
                {..}
            ]
        }
    """

    RINEX_HEADER_CLASS = RinexObsHeader

    def __init__(self, **kwargs):
        self.header = self.RINEX_HEADER_CLASS()
        self.interval_filter: int = kwargs.get("interval_filter", 0)
        self.backup_epochs = []
        self.rinex_obs_file = kwargs.get("rinex_obs_file", "")
        self.rinex_epochs: List[RinexEpoch] = kwargs.get("rinex_epochs", [])
        self.rinex_date = kwargs.get("rinex_date", datetime.datetime.now().date())
        self.filter_on_read: bool = kwargs.get("filter_on_read", True)
        self.found_obs_types = {}

    @staticmethod
    def get_start_time(file_sequence):
        """ """
        if file_sequence == "0":
            return datetime.time(0, 0)
        return datetime.time(ord(file_sequence.lower() - 97), 0)

    @staticmethod
    def get_epochs_possible(file_sequence, interval):
        """
        Get maximal epochs for rinex file sequence

        Args:
            file_sequence: str, [a-x0]
            interval: int, Rinex Epoch Interval

        Returns:
            int, Possible Epochs in File
        """
        ef = datetime.datetime.combine(
            datetime.date.today(), Rinex2ObsReader.get_start_time(file_sequence)
        )
        el = datetime.datetime.combine(
            datetime.date.today(), Rinex2ObsReader.get_end_time(file_sequence, interval)
        )
        return int((el - ef).total_seconds() / interval) + 1

    @staticmethod
    def prepare_line(line):
        new_line = line.replace("\r", "").replace("\n", "")
        if len(new_line) % 16 != 0:
            new_line += " " * (16 - len(new_line) % 16)
        return new_line

    @staticmethod
    def get_end_time(file_sequence, interval):
        """ """
        if file_sequence == "0":
            return datetime.time(23, 59, 60 - interval)
        return datetime.time(ord(file_sequence.lower() - 97), 59, 60 - interval)

    @staticmethod
    def is_valid_filename(filename, rinex_version=2):
        """
        Checks if filename is rinex conform
        """
        rinex_version = float(rinex_version)
        if (rinex_version < 3) & (rinex_version >= 2):
            filename_regex = Rinex2ObsReader.RINEX_FILE_NAME_REGEX
        elif rinex_version >= 3:
            filename_regex = Rinex3ObsReader.RINEX_FILE_NAME_REGEX
        else:
            return False
        return re.match(filename_regex, filename) is not None

    def set_rinex_obs_file(self, rinex_obs_file: str):
        raise NotImplementedError

    def correct_year2(self, year2):
        """
        Accourding to the RINEX Manual 2.10, chapter "6.5 2-digit Years"
        """
        if year2 < 80:
            return year2 + 2000
        else:
            return year2 + 1900

    # def do_thinning(self, interval):
    #     """ """
    #     thinned_epochs = [
    #         epoch
    #         for epoch in self.rinex_epochs
    #         if epoch.get_day_seconds() % interval == 0
    #     ]
    #     if len(self.backup_epochs) == 0:
    #         self.backup_epochs = self.rinex_epochs
    #     self.rinex_epochs = thinned_epochs

    # def undo_thinning(self):
    #     """ """
    #     self.rinex_epochs = self.backup_epochs
    #     self.backup_epochs = []

    def to_rinex2(self):
        """ """
        out = ""
        for rinex_epoch in self.rinex_epochs:
            out += "%s\n" % rinex_epoch.to_rinex2()
        return out

    def to_rinex3(self):
        """ """
        out = []
        for rinex_epoch in self.rinex_epochs:
            s = rinex_epoch.to_rinex3()
            out.append(s)
        return "\n".join(out)

    def read_header(self, sort_obs_types=True):
        """ """
        header = ""
        with open(self.rinex_obs_file, "r") as handler:
            for i, line in enumerate(handler):
                header += line
                if "END OF HEADER" in line:
                    break
        self.header = self.RINEX_HEADER_CLASS.from_header(header_string=header)
        for sat_sys in self.header.sys_obs_types:
            self.found_obs_types[sat_sys] = set()

    def add_satellite(self, satellite):
        """
        Adds satellite to satellite list if not already added

        Args:
            satellite: Satid as str regexp '[GR][ \\d]{2}'
        """
        if satellite not in self.header.satellites:
            self.header.satellites[satellite] = 0
        self.header.satellites[satellite] += 1

    def has_satellite_system(self, sat_sys):
        """
        Checks if Satellite Systems is present or not

        Args:
            sat_sys: str, Satellite System "GREJIS"

        Returns:
            bool, True, if Satellite System is present, else False
        """
        for epoch in self.rinex_epochs:
            if epoch.has_satellite_system(sat_sys):
                return True
        return False

    def update_header_obs(self):
        """
        Updates header information about first and last observation
        """

        # First and Last Observation
        self.header.first_observation = self.rinex_epochs[0].timestamp
        self.header.last_observation = self.rinex_epochs[-1].timestamp

    def read_satellite(self, sat_id, line):
        raise NotImplementedError

    def read_data_to_dict(self):
        raise NotImplementedError


class Rinex2ObsReader(RinexObsReader):
    """
    classdocs

    Args:
        datadict: {
            "epochs": [
                {
                    "id": "YYYY-mm-ddTHH:MM:SSZ",
                    "satellites": [
                        {
                            "id": "[GR][0-9]{2},
                            "observations": {
                                "[CLSPD][12]": {
                                    "value": ..,
                                    "lli": ..,
                                    "ss": ..
                                }
                            }{1,}
                        }
                    ]
                },
                {
                    "id": "..."
                    "satellites": [..]
                },
                {..}
            ]
        }
    """

    RINEX_HEADER_CLASS = Rinex2ObsHeader
    RINEX_FILE_NAME_REGEX = r"....\d\d\d[a-x0]\.\d\d[oO]"
    RINEX_FORMAT = 2
    RINEX_DATELINE_REGEXP = cc.RINEX2_DATELINE_REGEXP
    RINEX_DATELINE_REGEXP_SHORT = cc.RINEX2_DATELINE_REGEXP_SHORT
    RINEX_SATELLITES_REGEXP = cc.RINEX2_SATELLITES_REGEXP

    def __init__(self, **kwargs):
        """
        Constructor
        """
        super(Rinex2ObsReader, self).__init__(**kwargs)

    def set_rinex_obs_file(self, rinex_obs_file):
        self.rinex_obs_file = rinex_obs_file
        self.station_doy_session = os.path.basename(self.rinex_obs_file).split(".")[0]
        assert self.__class__.is_valid_filename(
            os.path.basename(self.rinex_obs_file), self.header.format_version
        )
        self.station = self.station_doy_session[:4]
        self.doy = int(self.station_doy_session[4:7])
        year2 = int(self.rinex_obs_file.split(".")[-1][:2])
        self.year = self.correct_year2(year2)

        self.rinex_file_sequence = self.station_doy_session[7]
        self.backup_epochs = []

    def read_satellite(self, sat_id, line):
        """
        Parses trough rnx observation and creates dict. Referring to the RINEX Handbook 2.10
        there are only up to 5 observation types per line. This method parses any line length

        Args:
            sat_id: str satellite number/name
            line: str rnx line containing observations
        Returns:
            dict: {sat_id: {otk1: otv1, otk2: otv2, ... otkn: otvn}}
        """

        sat_dict = {"id": sat_id, "observations": {}}
        for k in range(len(self.header.observation_types)):
            obs_type = self.header.observation_types[k]
            obs_col = line[(16 * k) : (16 * (k + 1))]
            obs_val = obs_col[:14].strip()

            if obs_val == "":
                obs_val = None
            else:
                float(obs_val)

            if len(obs_col) < 15:
                obs_lli = 0
            else:
                obs_lli = obs_col[14].strip()
                if obs_lli == "":
                    obs_lli = 0
                else:
                    obs_lli = int(obs_lli)

            if len(obs_col) < 16:
                obs_ss = 0
            else:
                obs_ss = obs_col[15].strip()
                if obs_ss == "":
                    obs_ss = 0
                else:
                    obs_ss = int(obs_ss)

            if obs_val is None:
                # Do not store empty obs_type
                continue

            sat_dict["observations"].update(
                {
                    obs_type + "_value": obs_val,
                    obs_type + "_lli": obs_lli,
                    obs_type + "_ss": obs_ss,
                }
            )
        return sat_dict

    def read_data_to_dict(self):
        """ """
        # SKIP HEADER
        with open(self.rinex_obs_file, "r") as handler:

            rinex_obs = {
                "epochs": [],
                "fileName": self.rinex_obs_file,
                "year4": self.year,
                "doy": self.doy,
                "markerName": self.header.marker_name,
                "epochInterval": self.header.interval,
                "epochFirst": None,
                "epochLast": None,
            }
            end_of_header = False
            while True:

                # Check for END_OF_FILE
                line = handler.readline()
                if "END OF HEADER" in line:
                    celery_logger.debug("End of Header Reached")
                    end_of_header = True
                if not end_of_header:
                    continue
                if line == "":
                    break

                # Get DateLine
                r = re.search(self.RINEX_DATELINE_REGEXP, line)
                if r is not None:
                    timestamp = datetime.datetime(
                        self.correct_year2(year2=int(r.group("year2"))),
                        int(r.group("month")),
                        int(r.group("day")),
                        int(r.group("hour")),
                        int(r.group("minute")),
                        int(float(r.group("second"))),
                    )
                    epoch = timestamp.strftime("%FT%TZ")

                    rnx_epoch = {
                        "id": epoch,
                        "satellites": [],
                    }
                    sats = r.group("sat1").strip()
                    # Number of Satellites
                    nos = int(r.group("nos"))
                    if nos == 0:
                        continue

                    additional_lines = int((nos - 1) / 12 % 12)
                    for j in range(additional_lines):
                        line = handler.readline()
                        r2 = re.search(self.RINEX_DATELINE_REGEXP_SHORT, line)
                        if r2 is not None:
                            sats += r2.group("sat2").strip()

                    # Get Observation Data
                    for j in range(nos):
                        # i += 1
                        sat_num = sats[(3 * j) : (3 * (j + 1))]
                        self.add_satellite(sat_num)

                        raw_obs = ""
                        for k in range(1 + int(len(self.header.observation_types) / 5)):
                            raw_obs = "%s%s" % (
                                raw_obs,
                                self.prepare_line(handler.readline()),
                            )

                        rnx_epoch["satellites"].append(
                            self.read_satellite(sat_id=sat_num, line=raw_obs)
                        )

                    # Sort Satellites within epoch
                    rnx_epoch["satellites"] = sorted(
                        rnx_epoch["satellites"], key=lambda sat: sat["id"]
                    )

                    rinex_obs["epochs"].append(rnx_epoch)
                    rinex_epoch = RinexEpoch(
                        timestamp=datetime.datetime.strptime(
                            rnx_epoch["id"], cc.RNX_FORMAT_DATETIME
                        ),
                        observation_types=self.header.observation_types,
                        satellites=rnx_epoch["satellites"],
                        rcv_clock_offset=self.header.rcv_clock_offset,
                    )
                    self.rinex_epochs.append(rinex_epoch)

            if len(rinex_obs["epochs"]) > 0:
                rinex_obs["epochFirst"] = rinex_obs["epochs"][0]["id"]
                rinex_obs["epochLast"] = rinex_obs["epochs"][-1]["id"]
            logger.debug("Successfully created data dict")


class Rinex3ObsReader(RinexObsReader):
    """
    classdocs

    Args:
        datadict: {
            "epochs": [
                {
                    "id": "YYYY-mm-ddTHH:MM:SSZ",
                    "satellites": [
                        {
                            "id": "[GR][0-9]{2},
                            "observations": {
                                "[CLSPD][1258][ACPQW]": {
                                    "value": ..,
                                    "lli": ..,
                                    "ss": ..
                                }
                            }{1,}
                        }
                    ]
                },
                {
                    "id": "..."
                    "satellites": [..]
                },
                {..}
            ]
        }
    """

    RINEX_FORMAT = 3
    RINEX_HEADER_CLASS = Rinex3ObsHeader
    RINEX_DATELINE_REGEXP = cc.RINEX3_DATELINE_REGEXP
    RINEX_DATELINE_REGEXP_SHORT = cc.RINEX3_DATELINE_REGEXP
    RINEX_SATELLITES_REGEXP = cc.RINEX3_SATELLITES_REGEXP
    RINEX_FILE_NAME_REGEX = cc.RINEX3_FORMAT_FILE_NAME

    def __init__(self, **kwargs):
        """
        Constructor, use the same as Rinex2ObsReader
        """
        super(Rinex3ObsReader, self).__init__(**kwargs)

    def set_rinex_obs_file(self, rinex_obs_file):
        self.rinex_obs_file = rinex_obs_file

        assert self.is_valid_filename(
            os.path.basename(self.rinex_obs_file), self.header.format_version
        )
        m = re.match(self.RINEX_FILE_NAME_REGEX, os.path.basename(self.rinex_obs_file))

        d = m.groupdict()
        self.station = d["station"]
        self.doy = int(d["doy"])
        self.year = int(d["year4"])
        self.file_period = d["file_period"]
        self.rinex_file_sequence = -1  # g[6]

        self.rinex_obs_file = rinex_obs_file

        self.backup_epochs = []

    @staticmethod
    def is_valid_filename(filename, rinex_version=3):
        """
        Checks if filename is rinex conform
        """
        rinex_version = float(rinex_version)
        if rinex_version >= 3.0:
            filename_regex = Rinex3ObsReader.RINEX_FILE_NAME_REGEX
        else:
            return False
        m = re.match(filename_regex, filename)
        return m is not None

    def read_data_to_dict(self):
        """ """
        # SKIP HEADER
        with open(self.rinex_obs_file, "r") as handler:

            logger.info("Parse Header")
            keep_running = True
            while keep_running:
                line = handler.readline()
                if "END OF HEADER" in line:
                    keep_running = False

            logger.info("Parse Epochs")
            keep_running = True
            while keep_running:
                # Check for END_OF_FILE
                line = handler.readline()
                if line == "":
                    keep_running = False
                    continue
                # Get DateLine
                r = line.startswith(">")
                if not r:
                    continue

                # > 2025 03 16 00 00  0.0000000  0 37
                # y = line[2:6]
                # m = line[7:9]
                # d = line[10:12]
                H = line[13:15]
                M = line[16:18]
                S = line[18:30]
                if self.interval_filter > 0:
                    sec_of_day = int(H) * 3600
                    sec_of_day = int(M) * 60
                    sec_of_day = int(float(S))
                    if sec_of_day % self.interval_filter != 0:
                        continue

                e = line[31]
                n = line[32:35]
                epoch = line[2:29]

                # logger.debug("Found Date")
                # timestamp = datetime.datetime(
                #     int(r.group("year4")),
                #     int(r.group("month")),
                #     int(r.group("day")),
                #     int(r.group("hour")),
                #     int(r.group("minute")),
                #     int(float(r.group("second"))),
                # )
                # epoch = timestamp.strftime("%FT%TZ")

                rnx_epoch = {
                    "id": epoch,
                    "satellites": [],
                }

                if e not in ["0", "1"]:
                    logger.info("Special event: {}".format(e))

                # Number of Satellites
                nos = int(n)

                for j in range(nos):
                    line = handler.readline()
                    epoch_sat = self.read_epoch_satellite(line)
                    if epoch_sat:
                        self.add_satellite("sat_num")
                        rnx_epoch["satellites"].append(epoch_sat["sat_data"])
                    else:
                        logger.warning("No Data")

                rinex_epoch = RinexEpoch(
                    # timestamp=datetime.datetime.strptime(
                    #     rnx_epoch["id"], cc.RNX_FORMAT_DATETIME
                    # ),
                    timestamp=epoch,
                    observation_types=self.header.sys_obs_types,
                    satellites=rnx_epoch["satellites"],
                    rcv_clock_offset=self.header.rcv_clock_offset,
                )
                self.rinex_epochs.append(rinex_epoch)
        logger.info("Successfully created data dict")

    def read_epoch_satellite(self, line):
        """ """
        # sat_data = re.search(cc.RINEX3_DATA_OBSEVATION_REGEXP, line)
        sat_num = line[:3]

        # Get Observation Data
        if sat_num is not None:
            # self.add_satellite(sat_num)
            return {
                "sat_num": sat_num,
                "sat_data": self.read_satellite(sat_id=sat_num, line=line),
            }
        return {}

    def read_satellite(self, sat_id, line):
        """
        Parses trough rnx observation and creates dict. Referring to the RINEX Handbook 3.03

        Args:
            sat_id: str satellite number/name
            line: str rnx line containing observations
        Returns:
            dict: {
                id: "sat_id",
                observations: {otk1: otv1, otk2: otv2, ... otkn: otvn}
            }
        """
        sat_dict = {"id": sat_id, "observations": {}}
        try:
            sat_sys = sat_id[0]
            chunk = line[3:]

            for obs_field in self.header.sys_obs_types[sat_sys]["obs_types"]:
                if not chunk:
                    break
                val = lli = ssi = None
                try:
                    val = chunk[:14]
                    chunk = chunk[14:]
                    lli = chunk[0]
                    chunk = chunk[1:]
                    ssi = chunk[0]
                    chunk = chunk[1:]
                except:
                    pass

                if val.strip() == "":
                    continue

                self.found_obs_types[sat_sys].add(obs_field)

                sat_dict["observations"][f"{obs_field}_value"] = float(val)

                sat_dict["observations"][f"{obs_field}_ssi"] = (
                    0 if not str(ssi).isnumeric() else int(ssi)
                )
                sat_dict["observations"][f"{obs_field}_lli"] = (
                    None if not str(lli).isnumeric() else int(lli)
                )

        except Exception as e:
            traceback.print_exc()
            raise (e)
        return sat_dict



"""
Created on Nov 7, 2016

@author: jurgen
"""

import datetime
import re
import abc
import traceback

from rinex_parser import constants as c
from rinex_parser.logger import logger


class RinexObsHeader(object):
    """ """

    __metaclass__ = abc.ABCMeta

    RE_HEADER_TYPES_OF_OBSERV = re.compile(c.RINEX2_HEADER_OBSERVATION_TYPES)
    RE_HEADER_OBS_DESCRIPTOR = c.RINEX2_HEADER_OBS_DESCRIPTOR

    def __init__(self, **kwargs):
        self.comment = kwargs.get("comment", "")
        self.format_version = float(kwargs.get("format_version"))
        self.file_type = kwargs.get("file_type", "OBSERVATION DATA")
        self.satellite_system = kwargs.get("satellite_system", "M (MIXED)")
        self.satellites = {}
        self.run_by = kwargs.get("run_by", "ASBRU")
        self.run_date = kwargs.get(
            "run_date", datetime.datetime.now().strftime("%FT%H:%MZ")
        )
        self.marker_name = kwargs.get("marker_name")
        self.marker_number = kwargs.get("marker_number")
        self.observer = kwargs.get("observer")
        self.agency = kwargs.get("agency")
        self.receiver_number = kwargs.get("receiver_number")
        self.receiver_type = kwargs.get("receiver_type")
        self.receiver_version = kwargs.get("receiver_version")
        self.antenna_number = kwargs.get("antenna_number")
        self.antenna_type = kwargs.get("antenna_type")
        self.approx_position_x = kwargs.get("approx_position_x", None)
        self.approx_position_y = kwargs.get("approx_position_y", None)
        self.approx_position_z = kwargs.get("approx_position_z", None)
        self.antenna_delta_height = kwargs.get("antenna_delta_height", None)
        self.antenna_delta_east = kwargs.get("antenna_delta_east", None)
        self.antenna_delta_north = kwargs.get("antenna_delta_north", None)
        self.wavelength_fact_l1 = kwargs.get("wavelength_fact_l1", None)
        self.wavelength_fact_l2 = kwargs.get("wavelength_fact_l2", None)
        self.wavelength_fact = kwargs.get("wavelength_fact", None)
        self.observation_types = kwargs.get("observation_types", [])
        self.interval = kwargs.get("interval", None)
        self.sampling = kwargs.get("sampling", None)
        self.first_observation:str = kwargs.get("first_observation", None)
        self.last_observation:str = kwargs.get("last_observation", None)
        self.time_system = kwargs.get("time_system", None)
        self.rcv_clock_offset = kwargs.get("rcv_clock_offset", None)
        self.leap_seconds = kwargs.get("leap_seconds", None)
        self.total_satellites = 0
        self.empty = ""
        self.sys_obs_types = {}  # {"G": {"obs_types": [..]}, "R": {...}, ...}
        self.other_headers = []

    @property
    def datadict(self) -> dict:
        d = {
            "marker_name": self.marker_name,
            "marker_number": self.marker_number,
            "sys_obs_types": self.sys_obs_types,
        }
        return d

    @classmethod
    def from_header(cls, header_string):
        """
        Args:
            header: str, RinexHeaderString
        Returns:
            RinexObsHeader (2 or 3)
        """
        tmp_header = cls()
        try:
            tmp_header.set_header(header_lines=header_string)
            return tmp_header
        except Exception as e:
            traceback.print_exc()
            logger.error(e)
            return None

    @staticmethod
    def parse_version_type(line):
        format_version = float(line[:9].strip())
        file_type = line[20:40].strip()
        satellite_system = line[40:60].strip()
        return {
            "format_version": format_version,
            "file_type": file_type,
            "satellite_system": satellite_system,
        }

    def to_rinex3(self, *args):
        raise NotImplementedError

    def set_interval(self, line):
        self.interval = int(float(line.split()[0].strip()))

    def set_version_type(self, line):
        d = self.parse_version_type(line)
        self.format_version = d["format_version"]
        self.file_type = d["file_type"]
        self.satellite_system = d["satellite_system"]

    def set_pgm_by_date(self, line):
        self.program = line[:20].strip()
        self.run_by = line[20:40].strip()
        self.run_date = line[40:60].strip()

    def set_comment(self, line):
        # new_comment = line[:60].strip()
        new_comment = line
        new_lines = []
        if self.comment == "":
            new_lines.append(
                f"{'RinexParser-1.1.1':20s}{'Asbru RiDaH':20s}{datetime.datetime.now().strftime(c.RNX_FORMAT_DATETIME_SHORT):15s} UTC COMMENT"
            )
            new_lines.append(new_comment)
            self.comment = "\n".join(new_lines)
        else:
            self.comment += "\n".join(new_lines)

    def set_marker_name(self, line):
        self.marker_name = line[:60].strip()

    def set_marker_number(self, line):
        self.marker_number = line[:20].strip()

    def set_observer_agency(self, line):
        self.observer = line[00:20].strip()
        self.agency = line[20:60].strip()
        # clean ADDREF issue
        self.agency = self.agency.split("     ")[0]

    def set_receiver(self, line):
        self.receiver_number = line[:20].strip()
        self.receiver_type = line[20:40].strip()
        self.receiver_version = line[40:60].strip()

    def set_antenna(self, line):
        self.antenna_number = line[:20].strip()
        self.antenna_type = line[20:40].strip()

    def set_approx_position(self, line):
        self.approx_position_x = float(line[00:14])
        self.approx_position_y = float(line[14:28])
        self.approx_position_z = float(line[28:42])

    def set_antenna_delta(self, line):
        self.antenna_delta_height = float(line[00:14])
        self.antenna_delta_east = float(line[14:28])
        self.antenna_delta_north = float(line[28:42])

    def set_wavelength_fact(self, line):
        self.wavelength_fact_l1 = int(line[00:6])
        self.wavelength_fact_l2 = int(line[6:12])
        wlf = line[12:18].strip()
        if not wlf:
            wlf = 0
        self.wavelength_fact = int(wlf)

    def set_observation_types(self, line):
        m = re.match(self.RE_HEADER_TYPES_OF_OBSERV, line)
        if m is None:
            return
        d = m.groupdict()
        if d["noo"].strip() != "":
            self.observation_types_number = int(d["noo"])
        observation_types = re.finditer(self.RE_HEADER_OBS_DESCRIPTOR, line)
        if self.observation_types is None:
            self.observation_types = []
        self.observation_types += observation_types

    def set_first_observation(self, line):
        self.first_observation = line[:43]
        self.time_system = line[43 + 5 : 43 + 5 + 3]

    def set_last_observation(self, line):
        self.last_observation = line[:43]

    def set_rcv_clock_offset(self, line):
        self.rcv_clock_offset = int(line[:6])

    def set_leap_seconds(self, line):
        self.leap_seconds = int(line[:6])

    def set_total_satellites(self, line):
        self.total_satellites = int(line[:6])

    def set_header(self, header_lines):
        """
        Args:
            header_lines: str, Lines are separated by "\n"
        """
        for line in header_lines.split("\n"):
            if line == "":
                continue

            header_label = line[60:]

            # logger.debug(line)

            if "END OF HEADER" in header_label:
                break

            # RINEX VERSION / TYPE
            elif "RINEX VERSION / TYPE" in header_label:
                self.set_version_type(line)
                # logger.debug("Rinex Version: '%f'" % self.format_version)

            elif "PGM / RUN BY / DATE" in header_label:
                self.set_pgm_by_date(line)

            elif "COMMENT" in header_label:
                self.set_comment(line)

            elif "MARKER NAME" in header_label:
                self.set_marker_name(line)

            elif "MARKER NUMBER" in header_label:
                self.set_marker_number(line)

            elif "OBSERVER / AGENCY" in header_label:
                self.set_observer_agency(line)

            elif "REC # / TYPE / VERS" in header_label:
                self.set_receiver(line)

            elif "ANT # / TYPE" in header_label:
                self.set_antenna(line)

            elif "APPROX POSITION XYZ" in header_label:
                self.set_approx_position(line)

            elif "ANTENNA: DELTA H/E/N" in header_label:
                self.set_antenna_delta(line)

            elif "WAVELENGTH FACT L1/2" in header_label:
                self.set_wavelength_fact(line)

            elif "# / TYPES OF OBSERV" in header_label:
                self.set_observation_types(line)

            elif "INTERVAL" in header_label:
                self.set_interval(line)

            elif "TIME OF FIRST OBS" in header_label:
                self.set_first_observation(line)

            elif "TIME OF LAST OBS" in header_label:
                self.set_last_observation(line)

            elif "RCV CLOCK OFFS APPL" in header_label:
                self.set_rcv_clock_offset(line)

            elif "LEAP SECONDS" in header_label:
                self.set_leap_seconds(line)

            elif "# OF SATELLITES" in header_label:
                self.set_total_satellites(line)

            else:
                # parse remaining headers
                self.other_headers.append(line)


class Rinex2ObsHeader(RinexObsHeader):

    def __init__(self, **kwargs):
        kwargs["format_version"] = kwargs.get("format_version", "2.11")
        super(Rinex2ObsHeader, self).__init__(**kwargs)
        self.rinex_export_version = 2

    def to_rinex3(self):
        """ """
        self.rinex_export_version = 3
        r3 = Rinex3ObsHeader(**self.__dict__)
        return r3.to_rinex3()

    def to_rinex2(self):
        """
        Return RINEX2 Header
        """
        self.rinex_export_version = 2
        r2h = """{format_version:9.2f}{empty:11s}{file_type:20s}{satellite_system:20s}RINEX VERSION / TYPE
{program:20s}{run_by:20s}{run_date:20s}PGM / RUN_BY / DATE
{comment}
{marker_name:60s}MARKER NAME
{marker_number:60s}MARKER NUMBER
{observer:20s}{agency:40s}OBSERVER / AGENCY
{receiver_number:20s}{receiver_type:20s}{receiver_version:20s}REC # / TYPE / VERS
{antenna_number:20s}{antenna_type:20s}{empty:20s}ANT # / TYPE
{approx_position_x:14.4f}{approx_position_x:14.4f}{approx_position_x:14.4f}{empty:18s}APPROX POSITION XYZ
{antenna_delta_height:14.4f}{antenna_delta_east:14.4f}{antenna_delta_north:14.4f}{empty:18s}ANTENNA: DELTA H/E/N
{wavelength_fact_l1:6d}{wavelength_fact_l2:6d}{wavelength_fact:6d}{empty:42s}WAVELENGTH FACT L1/L2
{interval:10.3f}{empty:50s}INTERVAL""".format(
            **self.__dict__
        )

        ots = self.get_observation_types()
        if ots:
            r2h += "\n%s" % ots

        if self.first_observation is not None:
            r2h += "\n{time_of_obs:30s}{empty:5s}{time_system:3s}{empty:9s}TIME OF FIRST OBS".format(
                time_of_obs=self.first_observation.strftime(c.RNX_FORMAT_OBS_TIME),
                empty="",
                time_system=self.time_system,
            )
        if self.last_observation is not None:
            r2h += "\n{time_of_obs:30s}{empty:5s}{time_system:3s}{empty:9s}TIME OF LAST OBS".format(
                time_of_obs=self.last_observation.strftime(c.RNX_FORMAT_OBS_TIME),
                empty="",
                time_system=self.time_system,
            )
        if self.rcv_clock_offset is not None:
            r2h += "\n{:6d}{:54s}RCV CLOCK OFFS APPL".format(self.rcv_clock_offset, "")
        if self.total_satellites > 0:
            r2h += "\n{:6d}{:54s}# OF SATELLITES"
        return "{}\n{:60s}END OF HEADER".format(r2h, "")

    def get_observation_types(self) -> str:
        ots = []
        s = ""
        for ot in self.observation_types:
            if ot == "C1C" and self.rinex_export_version == 2:
                ot = "C1"
            if ot in ["C2W", "C2P"] and self.rinex_export_version == 2:
                ot = "P2"
            if ot in ots:
                continue
            ots.append(ot)

        for i in range(len(ots)):
            ot = ots[i]
            if i % 9 == 0:
                if i == 0:
                    s = "{:6d}".format(len(ots))
                else:
                    s = "\n{:6s}".format("")
            s += "{:4s}{:2s}".format("", ot)
            if (i % 9 == 8) or (i == len(ots) - 1):
                s = "{:60s}# / TYPES OF OBSERV".format(s)
        return s


class Rinex3ObsHeader(Rinex2ObsHeader):

    RE_SYS_OBS_TYPE = re.compile(c.RINEX3_FORMAT_SYS_OBS_TYPES)
    RE_HEADER_OBS_DESCRIPTOR = c.RINEX3_HEADER_OBS_DESCRIPTOR

    def __init__(self, **kwargs):
        """ """
        kwargs["format_version"] = "3.03"
        super(Rinex3ObsHeader, self).__init__(**kwargs)
        self.rinex_export_version = 3
        self.last_sat_sys = ""

        self.marker_type = kwargs.get("marker_type", "NON_GEODETIC")

        self.antenna_delta_x = kwargs.get("antenna_delta_x", None)
        self.antenna_delta_y = kwargs.get("antenna_delta_y", None)
        self.antenna_delta_z = kwargs.get("antenna_delta_z", None)

        self.antenna_phasecenter = kwargs.get("antenna_phasecenter", None)

        self.antenna_b_sight_x = kwargs.get("b_sight_x", None)
        self.antenna_b_sight_y = kwargs.get("b_sight_y", None)
        self.antenna_b_sight_z = kwargs.get("b_sight_z", None)

        self.zerodir_azi = kwargs.get("zerodir_azi", None)
        self.zerodir_x = kwargs.get("zerodir_x", None)
        self.zerodir_y = kwargs.get("zerodir_y", None)
        self.zerodir_z = kwargs.get("zerodir_z", None)

        self.center_mass_x = kwargs.get("center_mass_x", None)
        self.center_mass_y = kwargs.get("center_mass_y", None)
        self.center_mass_z = kwargs.get("center_mass_z", None)

        self.sys_obs_types = kwargs.get("sys_obs_types", {})

        self.signal_strength_unit = kwargs.get("signal_strength_unit", None)

        self.sys_dcbs_applied = kwargs.get("sys_dcbs_applied", None)
        self.sys_pcvs_applied = kwargs.get("sys_pcvs_applied", None)
        self.sys_scale_factor = kwargs.get("sys_scale_factor", None)

        self.sys_phase_shift = kwargs.get("sys_phase_shift", None)

        self.glonass_slot_frq = kwargs.get("glonass_slot_frq", None)

        self.glonass_cod_phs_bis = kwargs.get("glonass_cod_phs_bis", None)

        self.leap_second_current = kwargs.get("leap_second_current", None)
        self.leap_second_other = kwargs.get("leap_second_other", None)
        self.leap_second_week = kwargs.get("leap_second_week", None)
        self.leap_second_day = kwargs.get("leap_second_day", None)
        self.leap_second_time_system = kwargs.get("leap_second_time_system", None)

        self.prn_obs = None

    def to_rinex3(self):
        self.rinex_export_version = 3
        rinex_header = [
            """{format_version:9.2f}{empty:11s}{file_type:20s}{satellite_system:20s}RINEX VERSION / TYPE
{program:20s}{run_by:20s}{run_date:20s}PGM / RUN_BY / DATE
{comment}
{marker_name:60s}MARKER NAME
{marker_number:60s}MARKER NUMBER
{marker_type:20s}{empty:40s}MARKER TYPE
{observer:20s}{agency:40s}OBSERVER / AGENCY
{receiver_number:20s}{receiver_type:20s}{receiver_version:20s}REC # / TYPE / VERS
{antenna_number:20s}{antenna_type:20s}{empty:20s}ANT # / TYPE
{approx_position_x:14.4f}{approx_position_y:14.4f}{approx_position_z:14.4f}{empty:18s}APPROX POSITION XYZ
{antenna_delta_height:14.4f}{antenna_delta_east:14.4f}{antenna_delta_north:14.4f}{empty:18s}ANTENNA: DELTA H/E/N""".format(
                **self.__dict__
            )
        ]
        adx = self.get_antenna_delta_xyz()
        if adx:
            rinex_header.append(adx)

        ots = self.get_observation_types().strip()
        if ots:
            rinex_header.append(ots)
            rinex_header.append(
                "{wavelength_fact_l1:6d}{wavelength_fact_l2:6d}{wavelength_fact:6d}{empty:42s}WAVELENGTH FACT L1/L2".format(
                    ots=self.get_observation_types(), **self.__dict__
                )
            )

        sot = self.get_sys_obs_types().strip()
        if sot:
            rinex_header.append(sot)

        smp = self.interval
        if self.sampling and self.sampling > 0:
            smp = self.sampling

        rinex_header.append(f"{smp:10.3f}{' ':50s}INTERVAL")

        if self.first_observation is not None:
            rinex_header.append(
                "{time_of_obs:30s}{empty:5s}{time_system:3s}{empty:9s}TIME OF FIRST OBS".format(
                    time_of_obs=self.first_observation,
                    empty="",
                    time_system=self.time_system,
                )
            )
        if self.last_observation is not None:
            rinex_header.append(
                "{time_of_obs:30s}{empty:5s}{time_system:3s}{empty:9s}TIME OF LAST OBS".format(
                    time_of_obs=self.last_observation,
                    empty="",
                    time_system=self.time_system,
                )
            )
        if self.rcv_clock_offset is not None:
            rinex_header.append(
                "{:6d}{:54s}RCV CLOCK OFFS APPL".format(self.rcv_clock_offset, "")
            )

        if self.total_satellites > 0:
            rinex_header.append(
                "{:6d}{:54s}# OF SATELLITES".format(self.total_satellites, "")
            )

        apc = self.get_antenna_phasecenter()
        if apc:
            rinex_header.append(apc)

        bsi = self.get_antenna_b_sight()
        if bsi:
            rinex_header.append(bsi)

        aza = self.get_antenna_zerodir_azi()
        if aza:
            rinex_header.append(aza)

        azx = self.get_antenna_zerodir_xyz()
        if azx:
            rinex_header.append(azx)

        com = self.get_center_mass_xyz()
        if com:
            rinex_header.append(com)

        if self.other_headers:
            rinex_header += self.other_headers

        # sot = self.get_sys_obs_types()
        # if sot:
        #     rinex_header.append(sot)

        rinex_header.append("{:60s}END OF HEADER".format(""))
        return "\n".join(rinex_header)

    def get_antenna_delta_xyz(self):
        """ """
        if (
            self.antenna_delta_x is None
            or self.antenna_delta_y is None
            or self.antenna_delta_z is None
        ):
            return None
        return "{:14.4f}{:14.4f}{:14.4f}{:18s}ANTENNA: DELTA X/Y/Z".format(
            self.antenna_delta_x, self.antenna_delta_y, self.antenna_delta_z, ""
        )

    def get_antenna_phasecenter(self):
        """ """
        if not self.antenna_phasecenter:
            return None
        apc = ""
        for ss in c.RINEX3_SATELLITE_SYSTEMS:
            if ss in self.antenna_phasecenter:
                temp = "{} {:3s}{:9.4f}{:14.4f}{:14.4f}".format(
                    ss,
                    self.antenna_phasecenter[ss]["observation_code"],
                    self.antenna_phasecenter[ss]["fixed_station_north"],
                    self.antenna_phasecenter[ss]["fixed_station_east"],
                    self.antenna_phasecenter[ss]["fixed_station_up"],
                )
                apc += "{:60s}ANTENNA: PHASECENTER" % temp
                if ss != "S":
                    apc += "\n"
        return apc

    def get_antenna_b_sight(self):
        if (
            self.antenna_b_sight_x is None
            or self.antenna_b_sight_y is None
            or self.antenna_b_sight_z is None
        ):
            return None
        return "{:14.4f}{:14.4f}{:14.4f}{:18s}ANTENNA: B.SIGHT XYZ".format(
            self.antenna_b_sight_x, self.antenna_b_sight_y, self.antenna_b_sight_z, ""
        )

    def get_antenna_zerodir_azi(self):
        if self.zerodir_azi is None:
            return None

        return "{:14.4}{:46s}ANTENNA: ZERODIR AZI".format(self.zerodir_azi, "")

    def get_antenna_zerodir_xyz(self):
        if self.zerodir_x is None or self.zerodir_y is None or self.zerodir_z is None:
            return None
        return "{:14.4f}{:14.4f}{:14.4f}{:18s}ANTENNA: ZERODIR XYZ".format(
            self.zerodir_x, self.zerodir_y, self.zerodir_z, ""
        )

    def get_center_mass_xyz(self):
        if (
            self.center_mass_x is None
            or self.center_mass_y is None
            or self.center_mass_z is None
        ):
            return None
        return "{:14.4f}{:14.4f}{:14.4f}{:18s}CENTER OF MASS: XYZ".format(
            self.center_mass_x, self.center_mass_y, self.center_mass_z, ""
        )

    def get_sys_obs_types(self):
        """ """
        if self.sys_obs_types is None:
            return None
        sot = []
        for sat_sys in c.RINEX3_SATELLITE_SYSTEMS:
            if sat_sys in self.sys_obs_types:
                temp = f"{sat_sys}  {len(self.sys_obs_types[sat_sys]['obs_types']):3d}"
                for i, v in enumerate(self.sys_obs_types[sat_sys]["obs_types"]):
                    temp += f" {v:3s}"
                    if ((i + 1) % 13 == 0) and (i > 0):
                        sot.append(f"{temp:60s}SYS / # / OBS TYPES")
                        temp = " " * 6
                sot.append(f"{temp:60s}SYS / # / OBS TYPES")
        return "\n".join(sot)

    def set_marker_type(self, line):
        self.marker_type = line[:20].strip()

    def set_antenna_delta_xyz(self, line):
        self.antenna_delta_x = float(line[00:14])
        self.antenna_delta_y = float(line[14:28])
        self.antenna_delta_z = float(line[28:42])

    def set_antenna_phasecenter(self, line):
        phase_center = {}
        satellite_system = line[0]
        phase_center["satellite_system"] = satellite_system
        phase_center["observation_code"] = line[2 : 2 + 3]
        phase_center["fixed_station_north"] = float(line[5:14])
        phase_center["fixed_station_east"] = float(line[14:28])
        phase_center["fixed_station_up"] = float(line[28:42])

        self.antenna_phasecenter = {satellite_system: phase_center}

    def set_antenna_b_sight_xyz(self, line):
        self.antenna_b_sight_x = float(line[00:14])
        self.antenna_b_sight_y = float(line[14:28])
        self.antenna_b_sight_z = float(line[28:42])

    def set_antenna_zerodir_azi(self, line):
        self.zerodir_azi = float(line[00:14])

    def set_antenna_zerodir_xyz(self, line):
        self.zerodir_x = float(line[00:14])
        self.zerodir_y = float(line[14:28])
        self.zerodir_z = float(line[28:42])

    def set_center_mass_xyz(self, line):
        self.center_mass_x = float(line[00:14])
        self.center_mass_y = float(line[14:28])
        self.center_mass_z = float(line[28:42])

    def set_sys_obs_types(self, line):
        result = self.RE_SYS_OBS_TYPE.match(line)

        d = result.groupdict()
        sat_sys = ""

        # line of SYS OBS TYPES
        if "sat_sys" in d and d["sat_sys"] is not None:
            sat_sys = d["sat_sys"]
            # No SYS OBS TYPE exists so far
            if sat_sys not in self.sys_obs_types:
                self.sys_obs_types[sat_sys] = {"obs_types": []}
            self.last_sat_sys = sat_sys
            # logger.info("Setting LastSatSys to {}".format(self.last_sat_sys))

        # Continuation line SYS OBS TYPES
        elif "obs_cont" in d and self.last_sat_sys:
            sat_sys = self.last_sat_sys

        # get all observation descriptors
        if sat_sys != "":
            obs_descriptors = re.finditer(c.RINEX3_FORMAT_OBS_DESCRIPTOR, line)
            for obs_descriptor in obs_descriptors:
                obs_descriptor_clean = obs_descriptor.group(0).strip()
                if obs_descriptor_clean not in self.sys_obs_types[sat_sys]["obs_types"]:
                    self.sys_obs_types[sat_sys]["obs_types"].append(
                        obs_descriptor_clean
                    )

        return sat_sys

    def set_signal_strength_unit(self, line):
        self.signal_strength_unit = line[:20].strip()

    def set_interval(self, line):
        self.interval = float(line[:10])

    def set_sys_dcbs_applied(self, line):
        satellite_system = line[0]
        dcbs = {
            "satellite_system": satellite_system,
            "program_name": line[2:19],
            "correction_url": line[19:60],
        }
        self.sys_dcbs_applied[satellite_system].update(dcbs)

    def set_sys_pcvs_applied(self, line):
        satellite_system = line[0]
        pcvs = {
            "satellite_system": satellite_system,
            "program_name": line[2:19],
            "correction_url": line[19:60],
        }
        self.sys_pcvs_applied[satellite_system].update(pcvs)

    def set_sys_scale_factor(self, line):
        satellite_system = line[0]
        nos = line[8:10].strip()
        if nos == "":
            nos = 0
        else:
            nos = int(nos)

        scale_factor = {
            "satellite_system": satellite_system,
            "scale_factor": int(line[2:6]),
            "involved_satellites_number": nos,
            "involved_satellites_list": [],
        }
        self.sys_scale_factor[satellite_system] = scale_factor
        self.set_sys_scale_factor_extra(line, satellite_system)

    def set_sys_scale_factor_extra(self, line, satellite_system):
        self.sys_scale_factor[satellite_system]["involved_satellites_list"] += line[
            10:58
        ].split()

    def set_phase_shift(self, line):
        satellite_system = line[0]
        carrier_phase = line[2:5]
        correction_applied = line[5:13].strip()
        if correction_applied == "":
            correction_applied = 0.0
        nis = 0
        if line[16:18].strip() != "":
            nis = int(line[16:18])
        phase_shift = {
            "satellite_system": satellite_system,
            "carrier_phase": carrier_phase,
            satellite_system: {
                carrier_phase: {
                    "correction_applied": correction_applied,
                    "involved_satellites_number": nis,
                    "involved_satellites_list": [],
                }
            },
        }
        if self.sys_phase_shift is None:
            self.sys_phase_shift = {}
        if satellite_system in self.sys_phase_shift:
            if carrier_phase in self.sys_phase_shift[satellite_system]:
                self.sys_phase_shift[satellite_system][carrier_phase].update(
                    phase_shift[satellite_system][carrier_phase]
                )
            else:
                self.sys_phase_shift[satellite_system][carrier_phase] = phase_shift[
                    satellite_system
                ][carrier_phase]
        else:
            self.sys_phase_shift[satellite_system] = phase_shift[satellite_system]

        self.set_phase_shift_extra(line, satellite_system, carrier_phase)
        return phase_shift

    def set_phase_shift_extra(self, line, satellite_system, carrier):
        self.sys_phase_shift[satellite_system][carrier][
            "involved_satellites_list"
        ] += line[18:58].split()

    def set_glonass_slot_frq(self, line):
        nos = int(line[0:3])
        self.glonass_slot_frq = {"satellites_number": nos, "satellites_list": {}}
        self.set_glonass_slot_frq_extra(line)

    def set_glonass_slot_frq_extra(self, line):
        for g in range(8):
            pos = g * 7
            pnr = line[4 + pos : 7 + pos].strip()
            if pnr == "":
                continue
            frq = int(line[8 + pos : 10 + pos])
            self.glonass_slot_frq["satellites_list"][pnr] = {"pnr": pnr, "frq": frq}

    def set_glonass_cod_phs_bis(self, line):
        for i in range(4):
            pos = 13 * i
            si = line[1 + pos : 4 + pos].strip()
            if si == "":
                continue
            if self.glonass_cod_phs_bis is None:
                self.glonass_cod_phs_bis = {}
            self.glonass_cod_phs_bis[si] = float(line[5 + pos : 13 + pos])

    def set_leap_seconds(self, line):
        time_system = "GPS"
        try:
            self.leap_second_current = int(line[0:6])
            self.leap_second_other = int(line[6:12])
            self.leap_second_week = int(line[12:18])
            self.leap_second_day = int(line[18:24])
            time_system = line[24:27].strip()
        except:
            time_system = "GPS"
        if time_system == "":
            time_system = "GPS"
        self.leap_second_time_system = time_system

    def set_header(self, header_lines):
        """
        Args:
            header_lines: str, Lines are separated by "\n"
        """
        # super(Rinex3ObsHeader, self).set_header(header_lines)
        last_sat_sys_ps = ""
        last_sat_sys_ps_cp = ""
        # last_sat_sys_ot = ""
        last_sat_sys_sf = ""

        for line in header_lines.split("\n"):
            if line == "":
                continue

            if "END OF HEADER" in line[60:]:
                break

            header_label = line[60:]

            if "RINEX VERSION / TYPE" in header_label:
                self.set_version_type(line)

            elif "PGM / RUN BY / DATE" in header_label:
                self.set_pgm_by_date(line)

            elif "COMMENT" in header_label:
                self.set_comment(line)

            elif "MARKER NAME" in header_label:
                self.set_marker_name(line)

            elif "MARKER NUMBER" in header_label:
                self.set_marker_number(line)

            elif "MARKER TYPE" in header_label:
                self.set_marker_type(line)

            elif "OBSERVER / AGENCY" in header_label:
                self.set_observer_agency(line)

            elif "REC # / TYPE / VERS" in header_label:
                self.set_receiver(line)

            elif "ANT # / TYPE" in header_label:
                self.set_antenna(line)

            elif "APPROX POSITION XYZ" in header_label:
                self.set_approx_position(line)

            elif "ANTENNA: DELTA H/E/N" in header_label:
                self.set_antenna_delta(line)

            elif "WAVELENGTH FACT L1/2" in header_label:
                self.set_wavelength_fact(line)

            elif "# / TYPES OF OBSERV" in header_label:
                self.set_observation_types(line)

            elif "INTERVAL" in header_label:
                self.set_interval(line)

            elif "TIME OF FIRST OBS" in header_label:
                self.set_first_observation(line)

            elif "TIME OF LAST OBS" in header_label:
                self.set_last_observation(line)

            elif "RCV CLOCK OFFS APPL" in header_label:
                self.set_rcv_clock_offset(line)

            elif "# OF SATELLITES" in header_label:
                self.set_total_satellites(line)

            elif "ANTENNA: DELTA X/Y/Z" in header_label:
                self.set_antenna_delta_xyz(line)

            elif "ANTENNA: PHASECENTER" in header_label:
                self.set_antenna_phasecenter(line)

            elif "ANTENNA: B.SIGHT XYZ" in header_label:
                self.set_antenna_b_sight_xyz(line)

            elif "ANTENNA: ZERODIR AZI" in header_label:
                self.set_antenna_zerodir_azi(line)

            elif "ANTENNA: ZERODIR XYZ" in header_label:
                self.set_antenna_zerodir_xyz(line)

            elif "CENTER OF MASS: XYZ" in header_label:
                self.set_center_mass_xyz(line)

            elif "SYS / # / OBS TYPES" in header_label:
                self.set_sys_obs_types(line)

            # elif 'SIGNAL STRENGTH UNIT' in header_label:
            #     self.set_signal_strength_unit(line)

            elif "SYS / DCBS APPLIED" in header_label:
                self.set_sys_dcbs_applied(line)

            elif "SYS / PCVS APPLIED" in header_label:
                self.set_sys_pcvs_applied(line)

            elif "SYS / SCALE FACTOR" in header_label and line[0] != " ":
                sf = self.set_sys_scale_factor(line)
                last_sat_sys_sf = sf["satellite_system"]

            # elif 'SYS / SCALE FACTOR' in header_label and line[0] == " ":
            #     self.set_sys_scale_factor_extra(line, last_sat_sys_sf)

            # elif 'SYS / PHASE SHIFT' in header_label and line[0] != " ":
            #     ps = self.set_phase_shift(line)
            #     last_sat_sys_ps = ps["satellite_system"]
            #     last_sat_sys_ps_cp = ps["carrier_phase"]

            # elif 'SYS / PHASE SHIFT' in header_label and line[0] == " ":
            #     self.set_phase_shift_extra(
            #         line, last_sat_sys_ps, last_sat_sys_ps_cp
            #     )

            # elif 'GLONASS SLOT / FRQ #' in header_label and line[0:3].strip() != "":
            #     self.set_glonass_slot_frq(line)

            # elif 'GLONASS SLOT / FRQ #' in header_label and line[0:3].strip() == "":
            #     self.set_glonass_slot_frq_extra(line)

            # elif 'GLONASS COD/PHS/BIS' in header_label:
            #     self.set_glonass_cod_phs_bis(line)

            # elif 'LEAP SECONDS' in header_label:
            #     self.set_leap_seconds(line)

            else:
                self.other_headers.append(line)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from math import trunc
from .utils import ceil, jwday, monthcalendarhelper
from . import gregorian

EPOCH = 1948320.5
WEEKDAYS = ("Yekshanbeh", "Doshanbeh",
            "Seshhanbeh", "Chaharshanbeh",
            "Panjshanbeh", "Jomeh", "Shanbeh")

HAS_31_DAYS = (1, 2, 3, 4, 5, 6)
HAS_30_DAYS = (7, 8, 9, 10, 11)


def leap(year):
    '''Is a given year a leap year in the Persian calendar ?'''
    if year > 0:
        y = 474
    else:
        y = 473

    return ((((((year - y % 2820) + 474) + 38) * 682) % 2816) < 682)


def to_jd(year, month, day):
    '''Determine Julian day from Persian date'''

    if year >= 0:
        y = 474
    else:
        y = 473
    epbase = year - y
    epyear = 474 + (epbase % 2820)

    if month <= 7:
        m = (month - 1) * 31
    else:
        m = (month - 1) * 30 + 6

    return day + m + trunc(((epyear * 682) - 110) / 2816) + (epyear - 1) * 365 + trunc(epbase / 2820) * 1029983 + (EPOCH - 1)


def from_jd(jd):
    '''Calculate Persian date from Julian day'''
    jd = trunc(jd) + 0.5

    depoch = jd - to_jd(475, 1, 1)
    cycle = trunc(depoch / 1029983)
    cyear = (depoch % 1029983)

    if cyear == 1029982:
        ycycle = 2820
    else:
        aux1 = trunc(cyear / 366)
        aux2 = cyear % 366
        ycycle = trunc(((2134 * aux1) + (2816 * aux2) + 2815) / 1028522) + aux1 + 1

    year = ycycle + (2820 * cycle) + 474

    if (year <= 0):
        year -= 1

    yday = (jd - to_jd(year, 1, 1)) + 1

    if yday <= 186:
        month = ceil(yday / 31)
    else:
        month = ceil((yday - 6) / 30)

    day = int(jd - to_jd(year, month, 1)) + 1

    return (year, month, day)


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def month_length(year, month):
    if month in HAS_30_DAYS or (month == 12 and leap(year)):
        return 30
    elif month in HAS_31_DAYS:
        return 31

    return 29

def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)
    return monthcalendarhelper(start_weekday, monthlen)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from calendar import isleap
from math import trunc

from . import gregorian, ordinal, julian
from .utils import jwday, n_weeks


MON = 0
TUE = 1
WED = 2
THU = 3
FRI = 4
SAT = 5
SUN = 6


def to_jd(year, week, day):
    '''Return Julian day count of given ISO year, week, and day'''
    return day + n_weeks(SUN, gregorian.to_jd(year - 1, 12, 28), week)


def from_jd(jd):
    '''Return tuple of ISO (year, week, day) for Julian day'''
    year = gregorian.from_jd(jd)[0]
    day = jwday(jd) + 1

    dayofyear = ordinal.from_jd(jd)[1]
    week = trunc((dayofyear - day + 10) / 7)

    # Reset year
    if week < 1:
        week = weeks_per_year(year - 1)
        year = year - 1

    # Check that year actually has 53 weeks
    elif week == 53 and weeks_per_year(year) != 53:
        week = 1
        year = year + 1

    return year, week, day


def weeks_per_year(year):
    '''Number of ISO weeks in a year'''
    # 53 weeks: any year starting on Thursday and any leap year starting on
    # Wednesday
    jan1 = jwday(gregorian.to_jd(year, 1, 1))

    if jan1 == THU or (jan1 == WED and isleap(year)):
        return 53
    else:
        return 52


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, week, day):
    return gregorian.from_jd(to_jd(year, week, day))


def format(year, week, day):
    return "{}-W{:02}-{}".format(year, week, day)



#!/usr/bin/python

import os
import datetime

from rinex_parser import constants as cc
from rinex_parser.logger import logger
from rinex_parser.ext import convertdate


class RinexQuality(object):
    """
    """
    RINEX_PERIOD_UNITS = {
        "D": 60*60*24,
        "H": 60*60,
        "M": 60,
    }

    def __init__(self, **kwargs):
        super(RinexQuality, self).__init__()
        self.rinex_format = kwargs.get("rinex_format", 3)

    def filter_by_observation_descriptor(self, epoch_satellites, observation_descriptor, satellite_system):
        """
        """
        for satellite in epoch_satellites:
            if "id" in satellite and \
                satellite["id"].startswith(satellite_system) and \
                "observations" in satellite:
                for observation in satellite["observations"]:
                    if observation.endswith("_value"):
                        if observation == observation_descriptor and \
                            satellite["observations"][observation] is not None:
                            yield satellite    
                        elif self.rinex_format == 3 and observation.startswith(observation_descriptor) and \
                            satellite["observations"][observation] is not None:
                            yield satellite
                            continue       

    def is_valid_epoch(self, epoch, satellite_systems=["G"], observation_descriptors=["L1", "L2"], satellites=5):
        """
        Checks if epoch suffices validity criterias. Per default these are:

        * Satellite System contains is GPS
        * At Least 5 Satellites within each Satellite System

        Returns: bool, True, if suffices criterias, else False
        """
        for observation_descriptor in observation_descriptors:
            for satellite_system in satellite_systems:
                i = 0
                i_test = self.filter_by_observation_descriptor(
                    epoch_satellites=epoch["satellites"], 
                    observation_descriptor=observation_descriptor, 
                    satellite_system=satellite_system
                )
                i += len(list(i_test))
                if i < satellites:
                    return False
        return True

    @staticmethod
    def get_datetime_utc(epoch_str):
        """
        """
        return datetime.datetime.strptime(epoch_str, cc.RNX_FORMAT_DATETIME)

    @staticmethod
    def get_session_code(second_of_day):
        """
        Args:
            second_of_day (int): Second of day (0..86399)
        Returns:
            str: The Hour Code (A..X)
        """
        i = int(second_of_day / 3600)
        return chr(i + 65)

    def do_prepare_datadict(self, datadict, gapsize):
        """

        """
        if "epochs" not in datadict or ("epochs" in datadict and len(datadict["epochs"]) == 0):
            logger.warn("No Epoch parsed")
            return ""

        for zeitpunkt in ["epochFirst", "epochLast"]:
            if (zeitpunkt not in datadict and \
                not isinstance(datadict[zeitpunkt], datetime.datetime)) or \
                (zeitpunkt in datadict and datadict[zeitpunkt] is None):
                datadict[zeitpunkt] = datetime.datetime.now().strftime(cc.RNX_FORMAT_DATETIME)

        dt0 = convertdate.convertdate.year_doy.to_datetime(datadict["year4"], datadict["doy"])

        if "epochPeriod" not in datadict:
            datadict["epochPeriod"] = "01D"  # Daily Files as default

        period_count = int(datadict["epochPeriod"][:2])
        period_unit = datadict["epochPeriod"][-1]

        period_seconds = int(self.RINEX_PERIOD_UNITS[period_unit] * period_count / datadict["epochInterval"])
        
        chkdoy = {
            "filename": os.path.basename(datadict["fileName"]),
            "station": datadict["markerName"],
            "year": datadict["year4"],
            "doy": datadict["doy"],
            "dom": dt0.day,
            "month": dt0.month,
            "gaps": [],
            "epoch_interval": int(datadict["epochInterval"]),
            "epochs_valid": 0,
            "epochs_max": int(period_seconds),
            "epochs_missing": 0,
            "epoch_first": datadict["epochFirst"],
            "epoch_last": datadict["epochLast"],
        }

        # logger.debug("Prepare")
        dt_epoch_first = datetime.datetime.strptime(chkdoy["epoch_first"], cc.RNX_FORMAT_DATETIME)
        dt_epoch_last = datetime.datetime.strptime(chkdoy["epoch_last"], cc.RNX_FORMAT_DATETIME)
        total_seconds = int((
            dt_epoch_last - dt_epoch_first
        ).total_seconds())

        # logger.debug("Filter")
        epoch_valid = []
        for epoch in datadict["epochs"]:
            if not self.is_valid_epoch(epoch):
                continue
            if epoch["id"] not in epoch_valid:
                epoch_valid.append(epoch["id"])

        # logger.debug("Append")
        chkdoy["epochs_valid"] = len(epoch_valid)
        # chkdoy["epochs_missing"] = chkdoy["epochs_max"] - chkdoy["epochs_valid"]
        epochs = []
        for epoch in epoch_valid:
            temp_utc = self.get_datetime_utc(epoch_str=epoch)
            if temp_utc not in epochs:
                epochs.append(temp_utc)
        # logger.debug("Sort")
        epochs = sorted(epochs)

        # logger.debug("Delta")
        gaps_less = 0
        gaps_more = 0
        for i in range(len(epochs) - 1):
            epoch_delta = (epochs[i + 1] - epochs[i]).total_seconds()
            if epoch_delta > chkdoy["epoch_interval"]:
                chkdoy["gaps"].append({
                    "gap_begin": epochs[i].strftime(cc.RNX_FORMAT_DATETIME),
                    "gap_end": epochs[i + 1].strftime(cc.RNX_FORMAT_DATETIME),
                    "gap_epoch_count": epoch_delta / chkdoy["epoch_interval"],
                    "gap_duration": epoch_delta,  # (gap_epoch_count * datadict["epochInterval"])
                })

                if epoch_delta <= gapsize * chkdoy["epoch_interval"]:
                    gaps_less += 1
                else:
                    gaps_more += 1

        chkdoy.update({
            "gaps_less": gaps_less,
            "gaps_more": gaps_more,
            "gapsize": gapsize,
            "date": dt0.strftime(cc.RNX_FORMAT_DATE),
            "epoch_first": dt_epoch_first.strftime(cc.RNX_FORMAT_DATETIME),
            "epoch_last": dt_epoch_last.strftime(cc.RNX_FORMAT_DATETIME),
            "total_secs": int(total_seconds),
        })

        return chkdoy

    def get_rinex_availability_as_dict(self, datadict, gapsize=5):
        """
        Get RinexAvailability as dictionary
        """
        chkdoy = self.do_prepare_datadict(datadict, gapsize)
        rinex_v = []
        d = {
            "date": chkdoy["date"],
            "station_name": chkdoy["station"],
            "second_from": 0,
            "second_until": 1,
            "epoch_interval": chkdoy["epoch_interval"],
            "session_code": 'A',
            "is_online": 1
        }        
        window_list = []
        window = {"valid_from": chkdoy["epoch_first"], "valid_until": ""}
        for gap in chkdoy["gaps"]:
            window["valid_until"] = gap["gap_begin"]
            window_list.append(dict(**window))
            window["valid_from"] = gap["gap_end"]
        window["valid_until"] = chkdoy["epoch_last"]
        window_list.append(dict(**window))
        
        for w in window_list:
            w_from = datetime.datetime.strptime(w["valid_from"], cc.RNX_FORMAT_DATETIME)
            w_from_c = datetime.datetime(w_from.year, w_from.month, w_from.day)
            w_until = datetime.datetime.strptime(w["valid_until"], cc.RNX_FORMAT_DATETIME)
            w_until_c = datetime.datetime(w_until.year, w_until.month, w_until.day)
            w_delta = w_until - w_from
            w_delta_hours = w_delta.total_seconds()/3600.0
            d["second_from"] = int((w_from - w_from_c).total_seconds())
            # split by hours, not tested for bigger than daily files
            i = 0
            if w_delta_hours > 1:
                for i in range(int(w_delta_hours)):
                    d["second_until"] = int((w_from.hour + i+1)*3600 - chkdoy["epoch_interval"])
                    # rinex_v_i = "{date};{station_name};{second_from};{second_until};{epoch_interval};{session_code};{is_online}".format(
                    #     **d
                    # )
                    # rinex_v.append(rinex_v_i)
                    d["second_from"] = int(d["second_until"] + chkdoy["epoch_interval"])
                    d["session_code"] = self.get_session_code(d["second_from"])
                    rinex_v.append(d)
            else:
                d["second_until"] = int((w_until - w_until_c).total_seconds())
                d["session_code"] = self.get_session_code(d["second_from"])  
                rinex_v.append(d)
        return rinex_v

    def get_rinex_availability_as_str(self, availability_dict):
        """
        """
        rinex_v = []
        for d in availability_dict:        
            rinex_v_i = "{date};{station_name};{second_from};{second_until};{epoch_interval};{session_code};{is_online}".format(
            **d
            )
            rinex_v.append(rinex_v_i)
        return "\n".join(rinex_v)

    def get_rinex_availability(self, datadict, gapsize=5):
        """
        Should generate an output like:

        'YYYY-MM-DD;STATION;SECOND_BEGIN;SECOND_END;EPOCH_INTERVAL;SESSION_CODE;IS_ONLINE'
        """
        rinex_v_dict = self.get_rinex_availability_as_dict(datadict, gapsize)
        return self.get_rinex_availability_as_str(rinex_v_dict)

    def get_rinstat_as_dict(self, datadict, gapsize=5):
        """
        Get RinstatData as a dictionary
        """
        chkdoy = self.do_prepare_datadict(datadict, gapsize)
        chkdoy["gaps_count"] = len(chkdoy["gaps"])
        chkdoy["epoch_first"] = datetime.datetime.strptime(chkdoy["epoch_first"], cc.RNX_FORMAT_DATETIME).strftime(cc.RNX_FORMAT_TIME)
        chkdoy["epoch_last"] = datetime.datetime.strptime(chkdoy["epoch_last"], cc.RNX_FORMAT_DATETIME).strftime(cc.RNX_FORMAT_TIME)
        chkdoy["gaps_prepared"] = []
        for gap in chkdoy["gaps"]:
            gap_dict = {
                "ge": int(gap["gap_epoch_count"]) - 1,
                "gs": gap["gap_duration"] - chkdoy["epoch_interval"],
                "gf": datetime.datetime.strptime(gap["gap_begin"], cc.RNX_FORMAT_DATETIME) + datetime.timedelta(seconds=chkdoy["epoch_interval"]),
                "gu": datetime.datetime.strptime(gap["gap_end"], cc.RNX_FORMAT_DATETIME),
            }
            chkdoy["gaps_prepared"].append(gap_dict)
        return chkdoy

    def get_rinstat_as_str(self, rinstat_dict):
        """
        """
        rinstat_dict["gaps_list"] = ""
        for gap in rinstat_dict["gaps_prepared"]:
            gap_str = "--- GAP:  {gf} - {gu}  {gs:10.1f} s {ge:10d} e".format(
                **gap
            )
            rinstat_dict["gaps_list"] += "\n{}".format(gap_str)

        # Report
        return """+++ >>>   {filename}{gaps_list}
+++ RNX.SUM   {date} ({doy})   {station}   {epoch_first} - {epoch_last}   {total_secs} s   {epoch_interval} s {gaps_count:7d}
+++    #maxepo #aepoch #mepoch #gaps>{gapsize} #gaps<{gapsize}
+++      {epochs_max:5d}   {epochs_valid:5d}   {epochs_missing:5d}   {gaps_more:5d}   {gaps_less:5d}
+++ <<<""".format(**rinstat_dict)


    def get_rinstat_out(self, datadict, gapsize=5):
        """
        Create Dataframe with filter:
        - more than 5 GPS per epoch
        - having at least L1 and L2

        old report format:

        +++ >>>   SSSSDOYS.YYo
        +++ RNX.SUM   YY MM DD (DOY)   SSSS   HH MM SS -  HH MM SS   DDDDD s   DD s
        +++    #maxepo  #epoch   #gaps #gaps>5 #gaps<5
        +++       DDDD    DDDD     DDD     DDD     DDD
        +++ <<<
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ 

        Values to Prepare
        * Filename
        * Date of File (YY MM DD)
        * Day of Year (DOY)
        * Stationname (SSSS)
        * Time of first epoch
        * Time of last epoch
        * Total seconds of file
        * epoch interval of file
        * Maximal possible Epochs
        * Actual amount of Epochs
        * Total number of Gaps in file
        * Number of Gaps greater than 5 epochs
        * Number of Gaps less-equal than 5 epochs 

        Args:
            gapsize: int, Amount of gaps to bother

        """
        chkdoy = self.get_rinstat_as_dict(datadict, gapsize)
        return self.get_rinstat_as_str(chkdoy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>

# source: Wikipedia
# month 13 is the Sansculottides
french_republican_days = {
    1: [
        u"Raisin",
        u"Safran",
        u"Châtaigne",
        u"Colchique",
        u"Cheval",
        u"Balsamine",
        u"Carotte",
        u"Amaranthe",
        u"Panais",
        u"Cuve",
        u"Pomme de terre",
        u"Immortelle",
        u"Potiron",
        u"Réséda",
        u"Âne",
        u"Belle de nuit",
        u"Citrouille",
        u"Sarrasin",
        u"Tournesol",
        u"Pressoir",
        u"Chanvre",
        u"Pêche",
        u"Navet",
        u"Amaryllis",
        u"Bœuf",
        u"Aubergine",
        u"Piment",
        u"Tomate",
        u"Orge",
        u"Tonneau",
    ],
    2: [
        u"Pomme",
        u"Céleri",
        u"Poire",
        u"Betterave",
        u"Oie",
        u"Héliotrope",
        u"Figue",
        u"Scorsonère",
        u"Alisier",
        u"Charrue",
        u"Salsifis",
        u"Mâcre",
        u"Topinambour",
        u"Endive",
        u"Dindon",
        u"Chervis",
        u"Cresson",
        u"Dentelaire",
        u"Grenade",
        u"Herse",
        u"Bacchante",
        u"Azerole",
        u"Garance",
        u"Orange",
        u"Faisan",
        u"Pistache",
        u"Macjonc",
        u"Coing",
        u"Cormier",
        u"Rouleau",
    ],
    3: [
        u"Raiponce",
        u"Turneps",
        u"Chicorée",
        u"Nèfle",
        u"Cochon",
        u"Mâche",
        u"Chou-fleur",
        u"Miel",
        u"Genièvre",
        u"Pioche",
        u"Cire",
        u"Raifort",
        u"Cèdre",
        u"Sapin",
        u"Chevreuil",
        u"Ajonc",
        u"Cyprès",
        u"Lierre",
        u"Sabine",
        u"Hoyau",
        u"Érable à sucre",
        u"Bruyère",
        u"Roseau",
        u"Oseille",
        u"Grillon",
        u"Pignon",
        u"Liège",
        u"Truffe",
        u"Olive",
        u"Pelle",
    ],
    4: [
        u"Tourbe",
        u"Houille",
        u"Bitume",
        u"Soufre",
        u"Chien",
        u"Lave",
        u"Terre végétale",
        u"Fumier",
        u"Salpêtre",
        u"Fléau",
        u"Granit",
        u"Argile",
        u"Ardoise",
        u"Grès",
        u"Lapin",
        u"Silex",
        u"Marne",
        u"Pierre à chaux",
        u"Marbre",
        u"Van",
        u"Pierre à plâtre",
        u"Sel",
        u"Fer",
        u"Cuivre",
        u"Chat",
        u"Étain",
        u"Plomb",
        u"Zinc",
        u"Mercure",
        u"Crible",
    ],
    5: [
        u"Lauréole",
        u"Mousse",
        u"Fragon",
        u"Perce-neige",
        u"Taureau",
        u"Laurier-thym",
        u"Amadouvier",
        u"Mézéréon",
        u"Peuplier",
        u"Coignée",
        u"Ellébore",
        u"Brocoli",
        u"Laurier",
        u"Avelinier",
        u"Vache",
        u"Buis",
        u"Lichen",
        u"If",
        u"Pulmonaire",
        u"Serpette",
        u"Thlaspi",
        u"Thimelé",
        u"Chiendent",
        u"Trainasse",
        u"Lièvre",
        u"Guède",
        u"Noisetier",
        u"Cyclamen",
        u"Chélidoine",
        u"Traîneau",
    ],
    6: [
        u"Tussilage",
        u"Cornouiller",
        u"Violier",
        u"Troène",
        u"Bouc",
        u"Asaret",
        u"Alaterne",
        u"Violette",
        u"Marceau",
        u"Bêche",
        u"Narcisse",
        u"Orme",
        u"Fumeterre",
        u"Vélar",
        u"Chèvre",
        u"Épinard",
        u"Doronic",
        u"Mouron",
        u"Cerfeuil",
        u"Cordeau",
        u"Mandragore",
        u"Persil",
        u"Cochléaria",
        u"Pâquerette",
        u"Thon",
        u"Pissenlit",
        u"Sylvie",
        u"Capillaire",
        u"Frêne",
        u"Plantoir",
    ],
    7: [
        u"Primevère",
        u"Platane",
        u"Asperge",
        u"Tulipe",
        u"Poule",
        u"Bette",
        u"Bouleau",
        u"Jonquille",
        u"Aulne",
        u"Couvoir",
        u"Pervenche",
        u"Charme",
        u"Morille",
        u"Hêtre",
        u"Abeille",
        u"Laitue",
        u"Mélèze",
        u"Ciguë",
        u"Radis",
        u"Ruche",
        u"Gainier",
        u"Romaine",
        u"Marronnier",
        u"Roquette",
        u"Pigeon",
        u"Lilas",
        u"Anémone",
        u"Pensée",
        u"Myrtille",
        u"Greffoir",
    ],
    8: [
        u"Rose",
        u"Chêne",
        u"Fougère",
        u"Aubépine",
        u"Rossignol",
        u"Ancolie",
        u"Muguet",
        u"Champignon",
        u"Hyacinthe",
        u"Râteau",
        u"Rhubarbe",
        u"Sainfoin",
        u"Bâton d'or",
        u"Chamerisier",
        u"Ver à soie",
        u"Consoude",
        u"Pimprenelle",
        u"Corbeille d'or",
        u"Arroche",
        u"Sarcloir",
        u"Statice",
        u"Fritillaire",
        u"Bourrache",
        u"Valériane",
        u"Carpe",
        u"Fusain",
        u"Civette",
        u"Buglosse",
        u"Sénevé",
        u"Houlette",
    ],
    9: [
        u"Luzerne",
        u"Hémérocalle",
        u"Trèfle",
        u"Angélique",
        u"Canard",
        u"Mélisse",
        u"Fromental",
        u"Martagon",
        u"Serpolet",
        u"Faux",
        u"Fraise",
        u"Bétoine",
        u"Pois",
        u"Acacia",
        u"Caille",
        u"Œillet",
        u"Sureau",
        u"Pavot",
        u"Tilleul",
        u"Fourche",
        u"Barbeau",
        u"Camomille",
        u"Chèvrefeuille",
        u"Caille-lait",
        u"Tanche",
        u"Jasmin",
        u"Verveine",
        u"Thym",
        u"Pivoine",
        u"Chariot",
    ],
    10: [
        u"Seigle",
        u"Avoine",
        u"Oignon",
        u"Véronique",
        u"Mulet",
        u"Romarin",
        u"Concombre",
        u"Échalote",
        u"Absinthe",
        u"Faucille",
        u"Coriandre",
        u"Artichaut",
        u"Girofle",
        u"Lavande",
        u"Chamois",
        u"Tabac",
        u"Groseille",
        u"Gesse",
        u"Cerise",
        u"Parc",
        u"Menthe",
        u"Cumin",
        u"Haricot",
        u"Orcanète",
        u"Pintade",
        u"Sauge",
        u"Ail",
        u"Vesce",
        u"Blé",
        u"Chalémie",
    ],
    11: [
        u"Épeautre",
        u"Bouillon blanc",
        u"Melon",
        u"Ivraie",
        u"Bélier",
        u"Prêle",
        u"Armoise",
        u"Carthame",
        u"Mûre",
        u"Arrosoir",
        u"Panic",
        u"Salicorne",
        u"Abricot",
        u"Basilic",
        u"Brebis",
        u"Guimauve",
        u"Lin",
        u"Amande",
        u"Gentiane",
        u"Écluse",
        u"Carline",
        u"Câprier",
        u"Lentille",
        u"Aunée",
        u"Loutre",
        u"Myrte",
        u"Colza",
        u"Lupin",
        u"Coton",
        u"Moulin",
    ],
    12: [
        u"Prune",
        u"Millet",
        u"Lycoperdon",
        u"Escourgeon",
        u"Saumon",
        u"Tubéreuse",
        u"Sucrion",
        u"Apocyn",
        u"Réglisse",
        u"Échelle",
        u"Pastèque",
        u"Fenouil",
        u"Épine vinette",
        u"Noix",
        u"Truite",
        u"Citron",
        u"Cardère",
        u"Nerprun",
        u"Tagette",
        u"Hotte",
        u"Églantier",
        u"Noisette",
        u"Houblon",
        u"Sorgho",
        u"Écrevisse",
        u"Bigarade",
        u"Verge d'or",
        u"Maïs",
        u"Marron",
        u"Panier",
    ],
    13: [
        u"La Fête de la Vertu",
        u"La Fête du Génie",
        u"La Fête du Travail",
        u"La Fête de l'Opinion",
        u"La Fête des Récompenses",
        u"La Fête de la Révolution",
    ]
}



import os
import argparse

from rinex_parser.obs_parser import RinexParser
from rinex_parser.logger import logger

parser = argparse.ArgumentParser()
parser.add_argument("finp")
parser.add_argument("--fout", type=str, default="")
parser.add_argument("--smp", type=int, default=0)
parser.add_argument("--country", type=str, default="XXX")
parser.add_argument("--rnx-version", type=int, choices=[2,3], default=3)


def run():
    args = parser.parse_args()
    assert os.path.exists(args.finp)
    
    rnx_parser = RinexParser(rinex_file=args.finp, rinex_version=args.rnx_version, sampling=args.smp)
    rnx_parser.run()

    if args.fout:
        out_file = args.fout
    else:
        out_file = os.path.join(
            os.path.dirname(args.finp),
            rnx_parser.get_rx3_long(country=args.country)
        )

    # Output Rinex File
    with open(out_file, "w") as rnx:
        logger.info(f"Write to file: {out_file}")
        rnx.write(rnx_parser.rinex_reader.header.to_rinex3())
        rnx.write("\n")
        rnx.write(rnx_parser.rinex_reader.to_rinex3())
        rnx.write("\n")

    logger.info("Done processing")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>

'''Convert to and from the Dublin day count'''

from . import daycount

EPOCH = 2415020  # Julian Day Count for Dublin Count 0

_dublin = daycount.DayCount(EPOCH)

to_gregorian = _dublin.to_gregorian

from_gregorian = _dublin.from_gregorian

to_jd = _dublin.to_jd

from_jd = _dublin.from_jd

from_julian = _dublin.from_julian

to_julian = _dublin.to_julian

to_datetime = _dublin.to_datetime

from_datetime = _dublin.from_datetime



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from math import trunc
from . import gregorian
from .utils import monthcalendarhelper, jwday

EPOCH = 2394646.5
EPOCH_GREGORIAN_YEAR = 1844

WEEKDAYS = ("Jamál", "Kamál", "Fidál", "Idál", "Istijlál", "Istiqlál", "Jalál")

MONTHS = ("Bahá", "Jalál", "Jamál", "‘Aẓamat", "Núr", "Raḥmat", "Kalimát", "Kamál", "Asmá’",
          "‘Izzat", "Mashíyyat", "‘Ilm", "Qudrat", "Qawl", "Masá’il", "Sharaf", "Sulṭán", "Mulk",
          "Ayyám-i-Há", "‘Alá")

ENGLISH_MONTHS = ("Splendor", "Glory", "Beauty", "Grandeur", "Light", "Mercy", "Words",
                  "Perfection", "Names", "Might", "Will", "Knowledge", "Power", "Speech", "Questions",
                  "Honour", "Sovereignty", "Dominion", "Days of Há", "Loftiness")


def to_jd(year, month, day):
    '''Determine Julian day from Bahai date'''
    gy = year - 1 + EPOCH_GREGORIAN_YEAR

    if month != 20:
        m = 0
    else:
        if gregorian.leap(gy + 1):
            m = -14
        else:
            m = -15
    return gregorian.to_jd(gy, 3, 20) + (19 * (month - 1)) + m + day


def from_jd(jd):
    '''Calculate Bahai date from Julian day'''

    jd = trunc(jd) + 0.5
    g = gregorian.from_jd(jd)
    gy = g[0]

    bstarty = EPOCH_GREGORIAN_YEAR

    if jd <= gregorian.to_jd(gy, 3, 20):
        x = 1
    else:
        x = 0
    # verify this next line...
    bys = gy - (bstarty + (((gregorian.to_jd(gy, 1, 1) <= jd) and x)))

    year = bys + 1
    days = jd - to_jd(year, 1, 1)
    bld = to_jd(year, 20, 1)

    if jd >= bld:
        month = 20
    else:
        month = trunc(days / 19) + 1
    day = int((jd + 1) - to_jd(year, month, 1))

    return year, month, day


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def month_length(year, month):
    if month == 19:
        if gregorian.leap(year + EPOCH_GREGORIAN_YEAR):
            return 5
        else:
            return 4

    else:
        return 19


def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)
    return monthcalendarhelper(start_weekday, monthlen)



'''
Created on Oct 25, 2016

@author: jurgen
'''

RINEX_SATELLITE_IDENTIFIER = r"(?P<satellite_system>[%s])(?P<satellite_number>[ \d]{2})"

RINEX2_SATELLITE_SYSTEMS = "GRE"
RINEX2_SATELLITES_REGEXP = RINEX_SATELLITE_IDENTIFIER % RINEX2_SATELLITE_SYSTEMS
RINEX2_DATELINE_REGEXP = r"^ (?P<year2>\d{2}) (?P<month>[ \d]{2}) (?P<day>[ \d]{2}) (?P<hour>[ \d]{2}) (?P<minute>[ \d]{2}) (?P<second>[ \d]{2}\.\d{7})  (?P<epoch_flag>[0-9])(?P<nos>[ \d]{3})(?P<sat1>(%s){0,12})(?P<clock_offset>.|[ \d]{2}\.\d{9})?" % RINEX2_SATELLITES_REGEXP
RINEX2_DATELINE_REGEXP_SHORT = r"^\ {32}(?P<sat2>(%s){,12})" % RINEX2_SATELLITES_REGEXP
RINEX2_DATA_OBSERVATION_DESCRIPTOR_REGEXP = r"(?P<value>[-\ \d]{10}\.\d{3})(?P<lli>[\ \d])()(?P<ss>[\ \d])"
RINEX2_DATA_OBSERVATION_REGEXP = r"^(%s|\ {16}){,5}" % RINEX2_DATA_OBSERVATION_DESCRIPTOR_REGEXP
RINEX2_HEADER_OBS_DESCRIPTOR = r"\ {4}[CPLDS][125678]"
RINEX2_HEADER_OBSERVATION_TYPES = r"^(?P<noo>[\ \d]{6})(?P<ots>(%s){0,9})" % RINEX2_HEADER_OBS_DESCRIPTOR

RINEX3_SATELLITE_SYSTEMS = "GREJCIS"
RINEX3_SATELLITES_REGEXP = RINEX_SATELLITE_IDENTIFIER % RINEX3_SATELLITE_SYSTEMS
RINEX3_DATELINE_REGEXP = r"^> (?P<year4>\d{4}) (?P<month>[ \d]{2}) (?P<day>[ \d]{2}) (?P<hour>[ \d]{2}) (?P<minute>[ \d]{2})(?P<second>[ \d]{3}.\d{7})  (?P<epoch_flag>\d)(?P<num_of_sats>[\d ]{3})((\ {6})(?P<rec_clock_offset>[\ \d-]{2}\.\d{12}))?"
RINEX3_MULTIPLE_OBS_REGEXP = r"(?P<first_o>([-\d\ ]{10}[\.\ ][\ \d]{3})([\d\ ])([\d\ ]))"
RINEX3_DATA_OBSEVATION_REGEXP = r"^(?P<sat_num>(%s))(?P<first_o>(([-\d\ ]{10}[\.\ ][\ \d]{3})([\d\ ])([\d\ ])))*(?P<last_o>([-\d\ ]{10}[\.\ ][\ \d]{3})([\d\ ])?([\d\ ])?)?" % RINEX3_SATELLITES_REGEXP
RINEX3_FORMAT_OBS_TIME = "%Y %m %d %H %M %S.0000000"
RINEX3_FORMAT_OBS_DESCRIPTOR = r"\ [CDLSX]\d[ABCDILMNPQSWX]?"
RINEX3_FORMAT_SYS_OBS_TYPES = r"^(((?P<sat_sys>[%s])  (?P<obs_amount>[\ \d]{3})|(?P<obs_cont>\ {6})))(?P<obs_descriptor>%s){0,13}.*(SYS / # / OBS TYPES)" % (RINEX3_SATELLITE_SYSTEMS, RINEX3_FORMAT_OBS_DESCRIPTOR)
RINEX3_FORMAT_FILE_NAME = r"(?P<station>\w{4})(\d)(\d)(?P<country>[A-Z]{3})_\w_(?P<year4>\d{4})(?P<doy>\d{3})(?P<hour>\d{2})(?P<minute>\d{2})_(?P<file_period>\d\d[A-Z])_((?P<data_freq>\d\d[A-Z])_)?\w\w.[rc]nx"
RINEX3_DATA_OBSERVATION_FIELD_REGEXP = r"((?P<value>\ *[\-\+]?[0-9]*\.?[0-9]{3})(?P<lli>[\ \d])?(?P<ssi>[\ \d])?)"
RINEX3_HEADER_OBS_DESCRIPTOR = r"\ {4}[CDLSX][125678][ABCDILMNPQSWX]"

RNX_FORMAT_DATETIME = "%Y-%m-%dT%H:%M:%SZ"
RNX_FORMAT_DATE = "%Y-%m-%d"
RNX_FORMAT_TIME = "%H:%M:%SZ"
RNX_FORMAT_DATETIME_SHORT = "%Y%m%d %H%M%S"
RNX_FORMAT_OBS_TIME = "  %Y    %m    %d    %H    %M   %S.%f0"

MARKER_TYPES = (
    "GEODETIC",         # Earth-fixed, high-precision monument
    "NON_GEODETIC",     # Earth-fixed, low-precision monument
    "NON_PHYSICAL",     # Generated from network processing
    "SPACEBORNE",       # Orbiting space vehicle
    "GROUND_CRAFT",     # Mobile terrestrial vehicle
    "WATER_CRAFT",      # Mobile water craft
    "AIRBORNE",         # Aircraft, ballon, etc.
    "FIXED_BUOY",       # "Fixed" on water surface
    "FLOATING_BUOY",    # Floating on water surface
    "FLOATING_ICE",     # Floatinc ice sheet, etc.
    "GLACIER",          # "Fixed" on a glacier
    "BALLISTIC",        # Rockets, shells, etc.
    "ANIMAL",           # Animal carrying a receiver
    "HUMAN"             # Human being
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from math import trunc
from .utils import ceil, jwday, monthcalendarhelper
from . import gregorian

EPOCH = 1948439.5
WEEKDAYS = ("al-'ahad", "al-'ithnayn",
            "ath-thalatha'", "al-'arb`a'",
            "al-khamis", "al-jum`a", "as-sabt")

HAS_29_DAYS = (2, 4, 6, 8, 10)
HAS_30_DAYS = (1, 3, 5, 7, 9, 11)


def leap(year):
    '''Is a given year a leap year in the Islamic calendar'''
    return (((year * 11) + 14) % 30) < 11


def to_jd(year, month, day):
    '''Determine Julian day count from Islamic date'''
    return (day + ceil(29.5 * (month - 1)) + (year - 1) * 354 + trunc((3 + (11 * year)) / 30) + EPOCH) - 1


def from_jd(jd):
    '''Calculate Islamic date from Julian day'''

    jd = trunc(jd) + 0.5
    year = trunc(((30 * (jd - EPOCH)) + 10646) / 10631)
    month = min(12, ceil((jd - (29 + to_jd(year, 1, 1))) / 29.5) + 1)
    day = int(jd - to_jd(year, month, 1)) + 1
    return (year, month, day)


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def month_length(year, month):
    if month in HAS_30_DAYS or (month == 12 and leap(year)):
        return 30

    return 29

def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)
    return monthcalendarhelper(start_weekday, monthlen)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file can be part of converdate.
# https://github.com/fitnr/convertdate

# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2018, jiargei <contact@messfuchs.at>

import datetime


def to_datetime(year, day_of_year):
    '''Creates a Datetime Object from Year and DayofYear. Time set to default'''
    return datetime.datetime(
        year, 1, 1) + datetime.timedelta(day_of_year - 1)


def from_datetime(date_time):
    '''Extracts Year and Day of Year out of a Datetime or Date object'''
    assert isinstance(date_time, (datetime.datetime, datetime.date))
    year = date_time.year
    day_of_year = date_time.timetuple().tm_yday
    return [year, day_of_year]



"""
Created on Nov 10, 2016

@author: jurgen
"""

import traceback
from rinex_parser import constants as cc
from rinex_parser.logger import logger

def ts_epoch_to_list(line: str) -> list:
    """Use epoch line and generate list of [y, m, d, H, M, S]."""
    y = int(line[2:6])
    m = int(line[7:9])
    d = int(line[10:12])
    H = int(line[13:15])
    M = int(line[16:18])
    S = float(line[18:30])
    return [y, m, d, H, M, S]


def ts_epoch_to_header(epoch: str) -> str:
    """Convert date from epoch format to header format."""
    # > 2025 03 16 00 00  0.0000000  0 37
    #   2025    03    17    19    00   00.0000000     GPS         TIME OF FIRST OBS
    line = f"> {epoch}"
    y, m, d, H, M, S = ts_epoch_to_list(line)
    s = f"  {y}    {m:02d}    {d:02d}    {H:02d}    {M:02d}  {S:11.7f}"
    return s

class RinexEpoch(object):
    """
    classdocs
    """

    def __init__(self, timestamp, observation_types, satellites, **kwargs):
        """
        Constructor
        Args:
            timestamp: datetime, Timestamp of epoch
            observation_types: list, List of observation types.
                               It's order will be used for export order
            satellites: dict, including the epoch's data
            epoch_flag: int, Epoch Flag (default 0)
            rcv_clock_offset: float, Offset of Receiver (default 0.0)
        """
        # self.timestamp: datetime.datetime = timestamp
        self.timestamp: str = timestamp
        self.observation_types = observation_types
        self.satellites = satellites
        self.epoch_flag = kwargs.get("epoch_flag", 0)
        self.rcv_clock_offset = kwargs.get("rcv_clock_offset", 0.0)

    def to_dict(self):
        d = {
            # "id": self.timestamp.strftime(cc.RNX_FORMAT_DATETIME),
            "id": self.timestamp,
            "satellites": self.satellites,
        }
        return d

    # def get_day_seconds(self):
    #     """
    #     :return: int, seconds till 00:00:00 of timestamp date
    #     """
    #     return (
    #         self.timestamp.second + self.timestamp.minute * 60 + self.timestamp * 3600
    #     )

    def is_valid(
        self,
        satellite_systems=["G"],
        observation_types=["L1", "L2", "L1C", "L1W"],
        satellites=5,
    ):
        """
        Checks if epoch suffices validity criterias. Per default these are:

        * Satellite System contains is GPS
        * Contains L1 and L2 observation Types
        * At Least 5 Satellites within each Satellite System

        Returns: bool, True, if suffices criterias, else False
        """
        for observation_type in observation_types:
            # logger.debug("Looking for Observation Type '%s'" % observation_type)
            for satellite_system in satellite_systems:
                # logger.debug("Looking for Satellite System '%s'" % satellite_system)
                i = 0
                for satellite in self.satellites:
                    if satellite["id"].startswith(satellite_system):
                        if (
                            observation_type in satellite["observations"]
                            and satellite["observations"][observation_type] is not None
                        ):
                            i += 1

                if i < satellites:
                    return False
        return True

    @staticmethod
    def get_val(val):
        try:
            if val is None:
                raise ValueError
            v = "{:14.3f}".format(float(val))
        except Exception as e:
            # logger.error(e)
            v = " " * 14
        return v

    @staticmethod
    def get_d(val):
        try:
            d = "{:d}".format(int(val))
            if d == "0":
                d = " "
        except Exception as e:
            # logger.error(e)
            d = " "
        return d

    def has_satellite_system(self, sat_sys):
        """
        Checks if Satellite Systems is present or not

        Args:
            sat_sys: str, Satellite System "GREJIS"

        Returns:
            bool, True, if Satellite System is present, else False
        """
        for sat in self.satellites:
            if sat.upper().startswith(sat_sys[0].upper()):
                return True
        return False

    def to_rinex2(self):
        """
        Exports Epoch with Rinex2 format

        Returns: str, Rinex2 Format
        """
        prn1 = ""
        prn2 = ""
        nos = len(self.satellites)
        data_lines = ""

        for i in range(nos):

            j = 0
            for ot in self.observation_types:
                j += 1
                if self.satellites[i]["observations"].has_key(ot):
                    val = self.get_val(
                        self.satellites[i]["observations"][ot + "_value"]
                    )
                    lli = self.get_d(self.satellites[i]["observations"][ot + "_lli"])
                    ss = self.get_d(self.satellites[i]["observations"][ot + "_ss"])

                    new_data = "{}{}{}".format(val, lli, ss)
                else:
                    new_data = " " * 16

                if ((j) % 5 == 0) and len(self.observation_types) > 5:
                    # logger.debug("New Data Line")
                    new_data = "%s\n" % new_data
                data_lines = "%s%s" % (data_lines, new_data)

            if i < nos - 1:
                data_lines += "\n"

            if i < 12:
                prn1 = "%s%s" % (prn1, self.satellites[i]["id"])
            else:
                if i % 12 == 0:
                    prn2 = "%s\n%s" % (prn2, " " * 32)
                prn2 = "%s%s" % (prn2, self.satellites[i]["id"])

        header_line = " {}  {:d}{:3d}{}{:12.9f}".format(
            # self.timestamp.strftime(cc.RINEX3_FORMAT_OBS_TIME),
            self.timestamp,
            self.epoch_flag,
            nos,
            prn1,
            self.rcv_clock_offset,
        )

        if prn2 != "":
            header_line = "%s%s" % (header_line, prn2)

        return "%s\n%s" % (header_line, data_lines)

    def get_satellite_systems(self):
        """
        Checks epoch for occuring satellite systems
        """
        satellite_systems = []
        for satellite_system in cc.RINEX3_SATELLITE_SYSTEMS:
            for satellite in self.satellites:
                if (
                    satellite["id"].startswith(satellite_system)
                    and satellite_system not in satellite_systems
                ):
                    satellite_systems.append(satellite_system)
        return satellite_systems

    def from_rinex2(self, rinex):
        """ """
        pass

    def to_rinex3(self):
        """
        Exports Epoch with Rinex3 format

        Returns: str, Rinex3 Format
        """
        nos = len(self.satellites)
        data_lines = ""

        rco = self.rcv_clock_offset if self.rcv_clock_offset else " "

        data_lines += (
            "> {epoch_time}  {epoch_flag}{nos:3d}{empty:6s}{rcvco}".format(
                epoch_time=self.timestamp,
                epoch_flag=self.epoch_flag,
                nos=nos,
                empty="",
                rcvco=rco,
            ).strip()
            + "\n"
        )

        # sort order

        sat_sys_order = "GRECJS"
        sat_sys_block = {}
        for sat_sys in sat_sys_order:
            sat_sys_block[sat_sys] = []

        # self.observation_types {"G": {"obs_types": [..]}, "R": {...}, ...}
        for _, item in enumerate(self.satellites):
            # item {"id": "G01", "observations": {"C1C_[value,lli,ss]": ...}}
            try:
                sat_sys = item["id"][0]  # G,R,E,C...
                obs_codes = self.observation_types[sat_sys]["obs_types"]
                new_sparse = ""
                new_data = ""
                for obs_code in obs_codes[::-1]:
                    try:
                        val = self.get_val(item["observations"][f"{obs_code}_value"])
                        lli = self.get_d(item["observations"][f"{obs_code}_lli"])
                        ssi = self.get_d(item["observations"][f"{obs_code}_ssi"])
                        if obs_code.startswith("L") and ssi != " " and lli == " ":
                            lli = "0"
                        new_part = f"{val}{lli}{ssi}"
                    except KeyError:
                        # Satellite does not have this code
                        new_part = " " * 16
                    except Exception as e:
                        traceback.print_exc()
                        new_part = " " * 16
                    finally:
                        if new_part.strip() != "":
                            new_sparse = " " * 16
                        else:
                            new_part = new_sparse
                        new_data = f"{new_part}{new_data}"

                new_data = f"{item['id']:3s}{new_data}"
                sat_sys_block[sat_sys].append(new_data)
            except Exception as e:
                print(e)

        sat_blocks = []
        for sat_sys in sat_sys_order:
            sat_blocks += sat_sys_block[sat_sys]
        data_lines += "\n".join(sat_blocks)

        return data_lines

    def from_rinex3(self, rinex):
        """ """



'''
__created__: 03. Okt. 2016

@author: friedrich
'''

import datetime

from .julian import from_datetime as julian_from_datetime
from .julian import from_jd as julian_from_jd

MON = 0
TUE = 1
WED = 2
THU = 3
FRI = 4
SAT = 5
SUN = 6

GPS_WEEK_DAYS = (
    MON, TUE, WED, THU, FRI, SAT, SUN
)


def from_jd(jd):
    '''Return tuple of ISO (year, week, day) for Julian day'''
    gpsweek = int((jd - 2444244.5) / 7)
    dow = (((jd - 2444244.5) / 7 - gpsweek) * 7)
    dow = int(round(dow * 86400) / 86400)
    return gpsweek, dow


def to_jd(gpsweek, day_of_week):
    return gpsweek * 7 + day_of_week + 2444244.5 + 13


def to_gpsweek_dos(gpsweek, day_of_week):
    jd = to_jd(gpsweek, day_of_week)
    return round(
        (((jd - 2444244.5) / 7 - gpsweek) * 7) * (24 * 60 * 60)
    )


def to_datetime(gpsweek, day_of_week):
    year, month, day = julian_from_jd(to_jd(gpsweek, day_of_week))
    return datetime.datetime(year, month, day)


def from_datetime(date_time):
    assert isinstance(date_time, datetime.datetime)
    return from_jd(julian_from_datetime(date_time))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>

# Most of this code is ported from Fourmilab's javascript calendar converter
# http://www.fourmilab.ch/documents/calendar/
# which was developed by John Walker
#
# The algorithms are believed to be derived from the following source:
# Meeus, Jean. Astronomical Algorithms . Richmond: Willmann-Bell, 1991. ISBN 0-943396-35-2.
#    The essential reference for computational positional astronomy.
#
from . import bahai
from . import daycount
from . import dublin
from . import gregorian
from . import hebrew
from . import holidays
from . import indian_civil
from . import islamic
from . import iso
from . import julian
from . import julianday
from . import mayan
from . import ordinal
from . import persian


__version__ = '2.0.7'

__all__ = [
    'holidays', 'bahai', 'dublin',
    'daycount',
    'french_republican', 'gregorian', 'hebrew',
    'indian_civil', 'islamic', 'iso',
    'julian', 'julianday',
    'mayan', 'persian', 'mayan',
    'ordinal',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
import time
import calendar
from math import trunc
from .utils import nth_day_of_month
from . import hebrew

# weekdays
MON = 0
TUE = 1
WED = 2
THU = 3
FRI = 4
SAT = 5
SUN = 6

# months
JAN = 1
FEB = 2
MAR = 3
APR = 4
MAY = 5
JUN = 6
JUL = 7
AUG = 8
SEP = 9
OCT = 10
NOV = 11
DEC = 12


def new_years(year):
    '''Jan 1st'''
    return (year, JAN, 1)


def martin_luther_king_day(year):
    '''third monday in January'''
    return nth_day_of_month(3, MON, JAN, year)


def lincolns_birthday(year):
    '''Feb 12'''
    return (year, FEB, 12)


def valentines_day(year):
    '''feb 14th'''
    return (year, FEB, 14)


def washingtons_birthday(year):
    '''Feb 22'''
    return (year, FEB, 22)


def presidents_day(year):
    '''3rd Monday of Feb'''
    return nth_day_of_month(3, MON, FEB, year)


def pulaski_day(year):
    '''1st monday in March'''
    return nth_day_of_month(1, MON, MAR, year)


def easter(year):
    '''Calculate western easter'''
    # formula taken from http://aa.usno.navy.mil/faq/docs/easter.html
    c = trunc(year / 100)
    n = year - 19 * trunc(year / 19)

    k = trunc((c - 17) / 25)

    i = c - trunc(c / 4) - trunc((c - k) / 3) + (19 * n) + 15
    i = i - 30 * trunc(i / 30)
    i = i - trunc(i / 28) * (1 - trunc(i / 28) * trunc(29 / (i + 1)) * trunc((21 - n) / 11))

    j = year + trunc(year / 4) + i + 2 - c + trunc(c / 4)
    j = j - 7 * trunc(j / 7)

    l = i - j

    month = 3 + trunc((l + 40) / 44)
    day = l + 28 - 31 * trunc(month / 4)

    return year, int(month), int(day)


def may_day(year):
    return (year, MAY, 1)


def mothers_day(year):
    '''2nd Sunday in May'''
    return nth_day_of_month(2, SUN, MAY, year)


def memorial_day(year):
    '''last Monday in May'''
    return nth_day_of_month(0, MON, MAY, year)


def fathers_day(year):
    '''3rd Sunday in June'''
    return nth_day_of_month(3, SUN, JUN, year)


def flag_day(year):
    '''June 14th'''
    return (year, JUN, 14)


def independence_day(year, observed=None):
    '''July 4th'''
    day = 4

    if observed:
        if calendar.weekday(year, JUL, 4) == SAT:
            day = 3

        if calendar.weekday(year, JUL, 4) == SUN:
            day = 5

    return (year, JUL, day)


def labor_day(year):
    '''first Monday in Sep'''
    return nth_day_of_month(1, MON, SEP, year)


def columbus_day(year, country='usa'):
    '''in USA: 2nd Monday in Oct
       Elsewhere: Oct 12'''
    if country == 'usa':
        return nth_day_of_month(2, MON, OCT, year)
    else:
        return (year, OCT, 12)


def halloween(year):
    '''Oct 31'''
    return (year, OCT, 31)


def election_day(year):
    '''1st Tues in Nov'''
    return nth_day_of_month(1, TUE, NOV, year)


def veterans_day(year):
    '''Nov 11'''
    return (year, NOV, 11)


def rememberance_day(year):
    return veterans_day(year)


def armistice_day(year):
    return veterans_day(year)


def thanksgiving(year, country='usa'):
    '''USA: last Thurs. of November, Canada: 2nd Mon. of October'''
    if country == 'usa':
        if year in [1940, 1941]:
            return nth_day_of_month(3, THU, NOV, year)
        elif year == 1939:
            return nth_day_of_month(4, THU, NOV, year)
        else:
            return nth_day_of_month(0, THU, NOV, year)

    if country == 'canada':
        return nth_day_of_month(2, MON, OCT, year)


def christmas_eve(year):
    '''24th of December'''
    return (year, DEC, 24)


def christmas(year):
    '''25th of December'''
    return (year, DEC, 25)


def new_years_eve(year):
    '''Dec 31st'''
    return (year, DEC, 31)

# Jewish holidays begins the sunset before the first (secular) day of the holiday
# With the eve option set, the day of this sunset is returned
# without the option, the (secular) day is returned


def hanukkah(year, eve=None):
    year, month, day = hebrew.to_jd_gregorianyear(year, hebrew.KISLEV, 25)
    if eve:
        day = day - 1
    return year, month, day


def rosh_hashanah(year, eve=None):
    year, month, day = hebrew.to_jd_gregorianyear(year, hebrew.TISHRI, 1)
    if eve:
        day = day - 1
    return year, month, day


def yom_kippur(year, eve=None):
    year, month, day = hebrew.to_jd_gregorianyear(year, hebrew.TISHRI, 10)
    if eve:
        day = day - 1

    return year, month, day


def passover(year, eve=None):
    year, month, day = hebrew.to_jd_gregorianyear(year, hebrew.NISAN, 15)
    if eve:
        day = day - 1

    return year, month, day

# Mexican holidays


def dia_constitucion(year, observed=True):
    if observed:
        return nth_day_of_month(1, MON, FEB, year)
    else:
        return (year, FEB, 5)


def natalicio_benito_juarez(year, observed=True):
    if observed:
        return nth_day_of_month(3, MON, MAR, year)
    else:
        return (year, MAR, 21)


def dia_independencia(year):
    return year, SEP, 16


def dia_revolucion(year):
    return nth_day_of_month(3, MON, NOV, year)


class Holidays(object):

    def __init__(self, year=None):
        self.year = year or time.localtime().tm_year

    def set_year(self, year):
        self.year = year

    def __repr__(self):
        return 'Holidays({})'.format(self.year)

    # the holidays...
    @property
    def christmas(self):
        return christmas(self.year)

    @property
    def christmas_eve(self):
        return christmas_eve(self.year)

    @property
    def thanksgiving(self, country='usa'):
        return thanksgiving(self.year, country)

    @property
    def new_years(self):
        return new_years(self.year)

    @property
    def new_years_eve(self):
        return new_years_eve(self.year)

    @property
    def independence_day(self):
        return independence_day(self.year, observed=True)

    @property
    def flag_day(self):
        return flag_day(self.year)

    @property
    def election_day(self):
        return election_day(self.year)

    @property
    def presidents_day(self):
        return presidents_day(self.year)

    @property
    def washingtons_birthday(self):
        return washingtons_birthday(self.year)

    @property
    def lincolns_birthday(self):
        return lincolns_birthday(self.year)

    @property
    def memorial_day(self):
        return memorial_day(self.year)

    @property
    def labor_day(self):
        return labor_day(self.year)

    @property
    def columbus_day(self, country='usa'):
        return columbus_day(self.year, country)

    @property
    def veterans_day(self):
        return veterans_day(self.year)

    @property
    def valentines_day(self):
        return valentines_day(self.year)

    @property
    def halloween(self):
        return halloween(self.year)

    @property
    def mothers_day(self):
        return mothers_day(self.year)

    @property
    def fathers_day(self):
        return fathers_day(self.year)

    @property
    def pulaski_day(self):
        return pulaski_day(self.year)

    @property
    def easter(self):
        return easter(self.year)

    @property
    def martin_luther_king_day(self):
        return martin_luther_king_day(self.year)

    @property
    def hanukkah(self):
        return hanukkah(self.year, eve=False)

    @property
    def rosh_hashanah(self):
        return rosh_hashanah(self.year, eve=False)

    @property
    def yom_kippur(self):
        return yom_kippur(self.year, eve=False)

    @property
    def passover(self):
        return passover(self.year, eve=False)

    @property
    def dia_constitucion(self):
        return dia_constitucion(self.year, observed=True)

    @property
    def natalicio_benito_juarez(self):
        return natalicio_benito_juarez(self.year, observed=True)

    @property
    def dia_independencia(self):
        return dia_independencia(self.year)

    @property
    def dia_revolucion(self):
        return dia_revolucion(self.year)


if __name__ == '__main__':
    holiday = Holidays(time.localtime().tm_year)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>

from datetime import datetime
from . import gregorian
from . import julian
from pytz import utc


def to_datetime(jdc):
    '''Return a datetime for the input floating point Julian Day Count'''
    year, month, day = gregorian.from_jd(jdc)

    # in jdc: 0.0 = noon, 0.5 = midnight
    # the 0.5 changes it to 0.0 = midnight, 0.5 = noon
    frac = (jdc + 0.5) % 1

    hours = int(24 * frac)

    mfrac = frac * 24 - hours
    mins = int(60 * round(mfrac, 6))

    sfrac = mfrac * 60 - mins
    secs = int(60 * round(sfrac, 6))

    msfrac = sfrac * 60 - secs

    # down to ms, which are 1/1000 of a second
    ms = int(1000 * round(msfrac, 6))

    return datetime(year, month, day, int(hours), int(mins), int(secs), int(ms), tzinfo=utc)


def from_datetime(dt):
    # take account of offset (if there isn't one, act like it's utc)
    try:
        dt = dt + dt.utcoffset()
    except TypeError:
        # Assuming UTC
        pass

    jdc = gregorian.to_jd(dt.year, dt.month, dt.day)

    hfrac = dt.hour / 24.
    mfrac = round(dt.minute / (24. * 60), 5)
    sfrac = round(dt.second / (24. * 60 * 60), 5)
    msfrac = dt.microsecond / (24. * 60 * 60 * 1000)

    return jdc + hfrac + mfrac + sfrac + msfrac


def to_gregorian(jdc):
    return gregorian.from_jd(jdc)


def from_gregorian(year, month, day):
    return gregorian.to_jd(year, month, day)


def to_julian(jdc):
    return julian.from_jd(jdc)


def from_julian(year, month, day):
    return julian.to_jd(year, month, day)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
import math
import calendar

TROPICALYEAR = 365.24219878  # Mean solar tropical year


def ceil(x):
    return int(math.ceil(x))


def floor(x):
    return int(math.floor(x))


def amod(a, b):
    '''Modulus function which returns numerator if modulus is zero'''
    modded = int(a % b)
    return b if modded is 0 else modded


# Sane people of the world, use calendar.weekday!
def jwday(j):
    '''Calculate day of week from Julian day'''
    return math.trunc((j + 0.5)) % 7


def weekday_before(weekday, jd):
    return jd - jwday(jd - weekday)


# @param weekday      Day of week desired, 0 = Monday
# @param jd           Julian date to begin search
# @param direction    1 = next weekday, -1 = last weekday
# @param offset       Offset from jd to begin search
def search_weekday(weekday, jd, direction, offset):
    '''Determine the Julian date for the next or previous weekday'''
    return weekday_before(weekday, jd + (direction * offset))


#  Utility weekday functions, just wrappers for search_weekday

def nearest_weekday(weekday, jd):
    return search_weekday(weekday, jd, 1, 3)


def next_weekday(weekday, jd):
    return search_weekday(weekday, jd, 1, 7)


def next_or_current_weekday(weekday, jd):
    return search_weekday(weekday, jd, 1, 6)


def previous_weekday(weekday, jd):
    return search_weekday(weekday, jd, -1, 1)


def previous_or_current_weekday(weekday, jd):
    return search_weekday(weekday, jd, 1, 0)


def n_weeks(weekday, jd, nthweek):
    j = 7 * nthweek

    if nthweek > 0:
        j += previous_weekday(weekday, jd)
    else:
        j += next_weekday(weekday, jd)

    return j


def monthcalendarhelper(start_weekday, month_length):
    end_weekday = start_weekday + (month_length - 1) % 7

    lpad = (start_weekday + 1) % 7
    rpad = (5 - end_weekday % 7) % 6

    days = [None] * lpad + list(range(1, 1 + month_length)) + rpad * [None]

    return [days[i:i + 7] for i in range(0, len(days), 7)]


def nth_day_of_month(n, weekday, month, year):
    '''Return (year, month, day) tuple that represents nth weekday of month in year. If n==0, returns last weekday of month. Weekdays: Monday=0'''
    if not (0 <= n <= 5):
        raise IndexError("Nth day of month must be 0-5. Received: {}".format(n))

    if not (0 <= weekday <= 6):
        raise IndexError("Weekday must be 0-6")

    firstday, daysinmonth = calendar.monthrange(year, month)

    # Get first WEEKDAY of month
    first_weekday_of_kind = 1 + (weekday - firstday) % 7

    if n == 0:
        # find last weekday of kind, which is 5 if these conditions are met, else 4
        if first_weekday_of_kind in [1, 2, 3] and first_weekday_of_kind + 28 < daysinmonth:
            n = 5
        else:
            n = 4

    day = first_weekday_of_kind + ((n - 1) * 7)

    if day > daysinmonth:
        raise IndexError("No {}th day of month {}".format(n, month))

    return (year, month, day)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of rinexparser.
# https://github.com/dach.pos/rinexparser

# Licensed under the Apache 2.0 license:
# http://opensource.org/licenses/apache2.0
# Copyright (c) 2018, jiargei <juergen.fredriksson@bev.gv.at>

import os
import math
import argparse
import datetime

from rinex_parser.constants import RNX_FORMAT_OBS_TIME
from rinex_parser.logger import logger
from rinex_parser.obs_factory import RinexObsFactory, RinexObsReader
from rinex_parser.obs_epoch import ts_epoch_to_header, ts_epoch_to_list


def run():
    parser = argparse.ArgumentParser(description="Analyse your Rinex files.")
    parser.add_argument("file", type=str, help="rinex file including full path")
    parser.add_argument("version", type=int, help="rinex version (2 or 3)")
    args = parser.parse_args()
    rinex_parser = RinexParser(rinex_version=args.version, rinex_file=args.file)
    rinex_parser.run()


class RinexParser:

    def __init__(self, rinex_file: str, rinex_version: int, *args, **kwargs):
        assert rinex_version in [
            2,
            3,
        ], f"Unknown version ({rinex_version} not in [2,3])"
        assert os.path.isfile(rinex_file), f"Not a File ({rinex_file})"
        self.rinex_version = rinex_version
        self.rinex_file = rinex_file
        if self.rinex_file != "":
            self.set_rinex_file(self.rinex_file)
        self.rinex_reader_factory = RinexObsFactory()
        self.rinex_reader: RinexObsReader = None
        self.filter_on_read: bool = kwargs.get("filter_on_read", True)
        self.sampling: int = kwargs.get("sampling", 0)
        self.__create_reader(self.rinex_version)
        self.rinex_reader.interval_filter = self.sampling

    @property
    def datadict(self):
        return self.get_datadict()
    
    def get_rx3_long(self, country: str = "XXX") -> str:
        code = self.rinex_reader.header.marker_name[:4]
        ts_f = ts_epoch_to_list("> " + self.rinex_reader.rinex_epochs[0].timestamp)
        ts_l = ts_epoch_to_list("> " + self.rinex_reader.rinex_epochs[-1].timestamp)
        ts_f[-1] = int(ts_f[-1])
        ts_l[-1] = int(ts_l[-1])
        dtF = datetime.datetime(*ts_f)
        dtL = datetime.datetime(*ts_l)
        dtD = (dtL - dtF).total_seconds()
        dtD_H = dtD / 3600.0
        dtD_D = dtD_H / 24.0
        dtD_W = dtD_D / 7.0
        unitPeriod = "H"
        unitCount = math.ceil(dtD_H)
        if dtD_H > 23:
            unitPeriod = "D"
            unitCount = math.ceil(dtD_D)
            if dtD_D > 7:
                unitPeriod = "W"
                unitCount = math.ceil(dtD_W)
        doy = int(dtF.strftime("%03j"))
        period = f"{unitCount:02d}{unitPeriod}"
        # c     c     y   j  h m
        # HKB200AUT_R_20250761900_01H_01S_MO.rnx
        # HKB200XXX_R_20250761900_01H_30S_MO.rnx
        smp = self.rinex_reader.header.interval
        if self.sampling > 0:
            smp = self.sampling
        smp = int(smp)
        return f"{code}00{country}_R_{dtF.year:04d}{doy:03d}{dtF.hour:02d}{dtF.minute:02d}_{period}_{smp:02d}S_MO.rnx"

    def get_datadict(self):
        d = {}
        d["epochs"] = [e.to_dict() for e in self.rinex_reader.rinex_epochs]
        d["epochInterval"] = self.rinex_reader.header.interval
        d["epochFirst"] = self.rinex_reader.rinex_epochs[0]["id"]
        d["epochLast"] = self.rinex_reader.rinex_epochs[-1]["id"]
        dtF = datetime.datetime.strptime(d["epochFirst"], RNX_FORMAT_OBS_TIME)
        dtL = datetime.datetime.strptime(d["epochLast"], RNX_FORMAT_OBS_TIME)
        dtD = (dtL - dtF).total_seconds()
        dtD_H = dtD / 3600.0
        dtD_D = dtD_H / 24.0
        dtD_W = dtD_D / 7.0
        unitPeriod = "H"
        unitCount = int(dtD_H)
        if dtD_H > 23:
            unitPeriod = "D"
            unitCount = int(dtD_D)
            if dtD_D > 7:
                unitPeriod = "W"
                unitCount = int(dtD_W)
        d["epochPeriod"] = f"{unitCount}{unitPeriod}"
        d["year4"] = dtF.strftime("%Y")
        d["doy"] = dtF.strftime("%j")
        d["markerName"] = self.rinex_reader.header.marker_name
        d["fileName"] = os.path.dirname(self.rinex_file)
        return d

    def __create_reader(self, rinex_version) -> RinexObsReader:
        self.rinex_reader = self.rinex_reader_factory.create_obs_reader_by_version(
            rinex_version
        )()

    def set_rinex_file(self, rinex_file):
        if os.path.isfile(rinex_file):
            self.rinex_file = rinex_file
        else:
            logger.warn("Could not find file: {}".format(rinex_file))
            self.rinex_file = ""

    def get_rinex_file(self):
        return self.rinex_file

    def do_create_datadict(self):
        """Read Rinex file and create datadict."""
        assert self.rinex_file != "", "Rinex file not specified"
        assert os.path.exists(
            self.rinex_file
        ), f"Could not find file ({self.rinex_file})"
        self.rinex_reader.set_rinex_obs_file(self.rinex_file)
        self.rinex_reader.read_header()
        # logger.info("done with header")
        self.rinex_reader.read_data_to_dict()

    def do_clear_datadict(self):
        """Read all epochs and find empty obs types and remove them from header."""
        found_obs_types = {}
        input_obs_types = {}
        for sat_sys in self.rinex_reader.header.sys_obs_types.keys():
            found_obs_types[sat_sys] = set()

        # go through all epochs and satellites and their obs types
        if not self.filter_on_read:
            for rinex_epoch in self.rinex_reader.rinex_epochs:
                for _, item in enumerate(rinex_epoch.satellites):
                    sat_sys = item["id"][0]  # GREC
                    for sat_obs in item["observations"].keys():
                        if (
                            sat_obs.endswith("_value")
                            and item["observations"][sat_obs] is not None
                        ):
                            obs_type = sat_obs.split("_")[0]
                            found_obs_types[sat_sys].add(obs_type)
        else:
            found_obs_types = self.rinex_reader.found_obs_types

        # go through all obs types from header and remove those who are not listed
        for sat_sys in self.rinex_reader.header.sys_obs_types:
            input_obs_types[sat_sys] = set(
                self.rinex_reader.header.sys_obs_types[sat_sys]["obs_types"]
            )
            for obs_type in input_obs_types[sat_sys] - found_obs_types[sat_sys]:
                logger.info(f"Remove unused OBS TYPE {sat_sys}-{obs_type}")
                self.rinex_reader.header.sys_obs_types[sat_sys]["obs_types"].remove(
                    obs_type
                )

    def run(self):
        assert os.path.isfile(self.rinex_file), f"Not a file ({self.rinex_file})"
        self.do_create_datadict()
        self.rinex_reader.header.first_observation = ts_epoch_to_header(self.rinex_reader.rinex_epochs[0].timestamp)
        self.rinex_reader.header.last_observation = ts_epoch_to_header(self.rinex_reader.rinex_epochs[-1].timestamp)



from .obs_reader import Rinex2ObsReader, Rinex3ObsReader, RinexObsReader
from .obs_header import Rinex2ObsHeader, Rinex3ObsHeader, RinexObsHeader


RINEX_CLASSES = {
    "versions": {
        "2": {
            "reader": Rinex2ObsReader,
            "header": Rinex2ObsHeader
        },
        "3": {
            "reader": Rinex3ObsReader,
            "header": Rinex3ObsHeader
        }
    }
}


class RinexObsFactory(object):
    """

    """

    def __create_obs_type_by_version(self, rinex_version, class_type):
        assert str(rinex_version) in RINEX_CLASSES["versions"]
        return RINEX_CLASSES["versions"][str(rinex_version)][class_type]

    def create_obs_reader_by_version(self, rinex_version) -> RinexObsReader:
        return self.__create_obs_type_by_version(rinex_version, "reader")

    def create_obs_header_by_version(self, rinex_version) -> RinexObsHeader:
        return self.__create_obs_type_by_version(rinex_version, "header")

    def __create_obs_type_by_file(self, rinex_file, class_type):
        with open(rinex_file, 'r') as handler:
            version_dict = RinexObsHeader.parse_version_type(handler.readline())
            version = int(version_dict["format_version"])
            return self.__create_obs_type_by_version(version, class_type)

    def create_obs_reader_by_file(self, rinex_file) -> RinexObsReader:
        return self.__create_obs_type_by_file(rinex_file, "reader")

    def create_obs_header_by_file(self, rinex_file) -> RinexObsHeader:
        return self.__create_obs_type_by_file(rinex_file, "header")




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from calendar import isleap as leap, monthrange
from .utils import floor, monthcalendarhelper, jwday

EPOCH = 1721425.5

INTERCALATION_CYCLE_YEARS = 400
INTERCALATION_CYCLE_DAYS = 146097

LEAP_SUPPRESSION_YEARS = 100
LEAP_SUPPRESSION_DAYS = 36524

LEAP_CYCLE_YEARS = 4
LEAP_CYCLE_DAYS = 1461

HAVE_30_DAYS = (4, 6, 9, 11)
HAVE_31_DAYS = (1, 3, 5, 7, 8, 10, 12)


def legal_date(year, month, day):
    '''Check if this is a legal date in the Gregorian calendar'''
    if month == 2:
        daysinmonth = 29 if leap(year) else 28
    else:
        daysinmonth = 30 if month in HAVE_30_DAYS else 31

    if not (0 < day <= daysinmonth):
        raise IndexError("Month {} doesn't have a day {}".format(month, day))

    return True


def to_jd2(year, month, day):
    '''Gregorian to Julian Day Count for years between 1801-2099'''
    # http://quasar.as.utexas.edu/BillInfo/JulianDatesG.html

    legal_date(year, month, day)

    if month <= 2:
        year = year - 1
        month = month + 12

    a = floor(year / 100)
    b = floor(a / 4)
    c = 2 - a + b
    e = floor(365.25 * (year + 4716))
    f = floor(30.6001 * (month + 1))
    return c + day + e + f - 1524.5


def to_jd(year, month, day):
    legal_date(year, month, day)

    if month <= 2:
        leap_adj = 0
    elif leap(year):
        leap_adj = -1
    else:
        leap_adj = -2

    return (
        EPOCH - 1 + (365 * (year - 1)) +
        floor((year - 1) / 4) + (-floor((year - 1) / 100)) +
        floor((year - 1) / 400) + floor((((367 * month) - 362) / 12) + leap_adj + day)
    )


def from_jd(jd):
    '''Return Gregorian date in a (Y, M, D) tuple'''
    wjd = floor(jd - 0.5) + 0.5
    depoch = wjd - EPOCH

    quadricent = floor(depoch / INTERCALATION_CYCLE_DAYS)
    dqc = depoch % INTERCALATION_CYCLE_DAYS

    cent = floor(dqc / LEAP_SUPPRESSION_DAYS)
    dcent = dqc % LEAP_SUPPRESSION_DAYS

    quad = floor(dcent / LEAP_CYCLE_DAYS)
    dquad = dcent % LEAP_CYCLE_DAYS

    yindex = floor(dquad / 365)
    year = (
        quadricent * INTERCALATION_CYCLE_YEARS +
        cent * LEAP_SUPPRESSION_YEARS +
        quad * LEAP_CYCLE_YEARS + yindex
    )

    if not (cent == 4 or yindex == 4):
        year += 1

    yearday = wjd - to_jd(year, 1, 1)

    isleap = leap(year)

    if yearday < 58 + isleap:
        leap_adj = 0
    elif isleap:
        leap_adj = 1
    else:
        leap_adj = 2

    month = floor((((yearday + leap_adj) * 12) + 373) / 367)
    day = int(wjd - to_jd(year, month, 1)) + 1

    return (year, month, day)


def month_length(year, month):
    return monthrange(year, month)[1]


def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)

    return monthcalendarhelper(start_weekday, monthlen)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from math import trunc
import itertools
from .utils import amod
from . import gregorian


EPOCH = 584282.5
HAAB_MONTHS = ["Pop", "Wo'", "Zip", "Sotz'", "Sek", "Xul",
               "Yaxk'in'", "Mol", "Ch'en", "Yax", "Sak'", "Keh",
               "Mak", "K'ank'in", "Muwan'", "Pax", "K'ayab", "Kumk'u", "Wayeb'"]

HAAB_TRANSLATIONS = [
    "Mat", "Frog", "Red", "Bat", "Bee", "Dog", "First Sun", "Water", "Cave", "Green",
    "White", "Red", "Encloser", "Yellow Sun", "Screech Owl", "Planting Time", "Turtle", "Ripe Corn", "Nameless"]

TZOLKIN_NAMES = ["Imix'", "Ik'", "Ak'b'al", "K'an", "Chikchan",
                 "Kimi", "Manik'", "Lamat", "Muluk", "Ok",
                 "Chuwen", "Eb'", "B'en", "Ix", "Men",
                 "K'ib'", "Kab'an", "Etz'nab'", "Kawak", "Ajaw"]

TZOLKIN_TRANSLATIONS = ['Water', 'Wind', 'Darkness', 'Net', 'Feathered Serpent',
                        'Death', 'Deer', 'Seed', 'Jade', 'Dog',
                        'Thread', 'Path', 'Maize', 'Tiger', 'Bird', 'Will',
                        'Wisdom', 'Obsidian Knife', 'Thunder', 'Sun']


def to_jd(baktun, katun, tun, uinal, kin):
    '''Determine Julian day from Mayan long count'''
    return EPOCH + (baktun * 144000) + (katun * 7200) + (tun * 360) + (uinal * 20) + kin


def from_jd(jd):
    '''Calculate Mayan long count from Julian day'''
    d = jd - EPOCH
    baktun = trunc(d / 144000)
    d = (d % 144000)
    katun = trunc(d / 7200)
    d = (d % 7200)
    tun = trunc(d / 360)
    d = (d % 360)
    uinal = trunc(d / 20)
    kin = int((d % 20))

    return (baktun, katun, tun, uinal, kin)


def to_gregorian(baktun, katun, tun, uinal, kin):
    jd = to_jd(baktun, katun, tun, uinal, kin)
    return gregorian.from_jd(jd)


def from_gregorian(year, month, day):
    jd = gregorian.to_jd(year, month, day)
    return from_jd(jd)


def to_haab(jd):
    '''Determine Mayan Haab "month" and day from Julian day'''
    # Number of days since the start of the long count
    lcount = trunc(jd) + 0.5 - EPOCH
    # Long Count begins 348 days after the start of the cycle
    day = (lcount + 348) % 365

    count = day % 20
    month = trunc(day / 20)

    return int(count), HAAB_MONTHS[month]


def to_tzolkin(jd):
    '''Determine Mayan Tzolkin "month" and day from Julian day'''
    lcount = trunc(jd) + 0.5 - EPOCH
    day = amod(lcount + 4, 13)
    name = amod(lcount + 20, 20)
    return int(day), TZOLKIN_NAMES[int(name) - 1]


def lc_to_haab(baktun, katun, tun, uinal, kin):
    jd = to_jd(baktun, katun, tun, uinal, kin)
    return to_haab(jd)


def lc_to_tzolkin(baktun, katun, tun, uinal, kin):
    jd = to_jd(baktun, katun, tun, uinal, kin)
    return to_tzolkin(jd)


def lc_to_haab_tzolkin(baktun, katun, tun, uinal, kin):
    jd = to_jd(baktun, katun, tun, uinal, kin)
    dates = to_tzolkin(jd) + to_haab(jd)
    return "{0} {1} {2} {3}".format(*dates)


def translate_haab(h):
    return dict(list(zip(HAAB_MONTHS, HAAB_TRANSLATIONS))).get(h)


def translate_tzolkin(tz):
    return dict(list(zip(TZOLKIN_NAMES, TZOLKIN_TRANSLATIONS))).get(tz)


def _haab_count(day, month):
    '''Return the count of the given haab in the cycle. e.g. 0 Pop == 1, 5 Wayeb' == 365'''
    if day < 0 or day > 19:
        raise IndexError("Invalid day number")

    try:
        i = HAAB_MONTHS.index(month)
    except ValueError:
        raise ValueError("'{0}' is not a valid Haab' month".format(month))

    return min(i * 20, 360) + day


def _tzolkin_from_count(count):
    number = amod(count, 13)
    name = TZOLKIN_NAMES[count % 20 - 1]
    return number, name


def _tzolkin_count(day, name):
    if day < 1 or day > 13:
        raise IndexError("Invalid day number")

    days = set(x + day for x in range(0, 260, 13))

    try:
        n = 1 + TZOLKIN_NAMES.index(name)
    except ValueError:
        raise ValueError("'{0}' is not a valid Tzolk'in day name".format(name))

    names = set(y + n for y in range(0, 260, 20))
    return days.intersection(names).pop()


def tzolkin_generator(number=None, name=None):
    '''For a given tzolkin name/number combination, return a generator
    that gives cycle, starting with the input'''

    # By default, it will start at the beginning
    number = number or 13
    name = name or "Ajaw"

    if number > 13:
        raise ValueError("Invalid day number")

    if name not in TZOLKIN_NAMES:
        raise ValueError("Invalid day name")

    count = _tzolkin_count(number, name)

    ranged = itertools.chain(list(range(count, 260)), list(range(1, count)))

    for i in ranged:
        yield _tzolkin_from_count(i)


def longcount_generator(baktun, katun, tun, uinal, kin):
    '''Generate long counts, starting with input'''
    j = to_jd(baktun, katun, tun, uinal, kin)

    while True:
        yield from_jd(j)
        j = j + 1


def next_haab(month, jd):
    '''For a given haab month and a julian day count, find the next start of that month on or after the JDC'''
    if jd < EPOCH:
        raise IndexError("Input day is before Mayan epoch.")

    hday, hmonth = to_haab(jd)

    if hmonth == month:
        days = 1 - hday

    else:
        count1 = _haab_count(hday, hmonth)
        count2 = _haab_count(1, month)

        # Find number of days between haab of given jd and desired haab
        days = (count2 - count1) % 365

    # add in the number of days and return new jd
    return jd + days


def next_tzolkin(tzolkin, jd):
    '''For a given tzolk'in day, and a julian day count, find the next occurrance of that tzolk'in after the date'''
    if jd < EPOCH:
        raise IndexError("Input day is before Mayan epoch.")

    count1 = _tzolkin_count(*to_tzolkin(jd))
    count2 = _tzolkin_count(*tzolkin)

    add_days = (count2 - count1) % 260
    return jd + add_days


def next_tzolkin_haab(tzolkin, haab, jd):
    '''For a given haab-tzolk'in combination, and a Julian day count, find the next occurrance of the combination after the date'''
    # get H & T of input jd, and their place in the 18,980 day cycle
    haabcount = _haab_count(*to_haab(jd))
    haab_desired_count = _haab_count(*haab)

    # How many days between the input day and the desired day?
    haab_days = (haab_desired_count - haabcount) % 365

    possible_haab = set(h + haab_days for h in range(0, 18980, 365))

    tzcount = _tzolkin_count(*to_tzolkin(jd))
    tz_desired_count = _tzolkin_count(*tzolkin)
    # How many days between the input day and the desired day?
    tzolkin_days = (tz_desired_count - tzcount) % 260

    possible_tz = set(t + tzolkin_days for t in range(0, 18980, 260))

    try:
        return possible_tz.intersection(possible_haab).pop() + jd
    except KeyError:
        raise IndexError("That Haab'-Tzolk'in combination isn't possible")


def month_length(month):
    """Not the actual length of the month, but accounts for the 5 unlucky/nameless days"""
    if month == "Wayeb'":
        return 5
    else:
        return 20


def haab_monthcalendar(baktun=None, katun=None, tun=None, uinal=None, kin=None, jdc=None):
    '''For a given long count, return a calender of the current haab month, divided into tzolkin "weeks"'''
    if not jdc:
        jdc = to_jd(baktun, katun, tun, uinal, kin)

    haab_number, haab_month = to_haab(jdc)
    first_j = jdc - haab_number + 1

    tzolkin_start_number, tzolkin_start_name = to_tzolkin(first_j)

    gen_longcount = longcount_generator(*from_jd(first_j))
    gen_tzolkin = tzolkin_generator(tzolkin_start_number, tzolkin_start_name)

    # 13 day long tzolkin 'weeks'
    lpad = tzolkin_start_number - 1
    rpad = 13 - (tzolkin_start_number + 19 % 13)

    monlen = month_length(haab_month)

    days = [None] * lpad + list(range(1, monlen + 1)) + rpad * [None]

    def g(x, generate):
        if x is None:
            return None
        return next(generate)

    return [[(k, g(k, gen_tzolkin), g(k, gen_longcount)) for k in days[i:i + 13]] for i in range(0, len(days), 13)]


def haab_monthcalendar_prospective(haabmonth, jdc):
    '''Give the monthcalendar for the next occurance of given haab month after jdc'''
    return haab_monthcalendar(jdc=next_haab(haabmonth, jdc))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from setuptools import setup

try:
    readme = open('README.rst').read()
except IOError:
    readme = ''

with open('convertdate/__init__.py') as i:
    version = next(r for r in i.readlines() if '__version__' in r).split('=')[1].strip('"\' \n')

setup(
    name="convertdate",

    version=version,

    description=("Converts between Gregorian dates and other calendar systems."
                 "Calendars included: Baha'i, French Republican, Hebrew, "
                 "Indian Civil, Islamic, ISO, Julian, Mayan and Persian."),

    long_description=readme,

    author="Neil Freeman",

    license='MIT',

    author_email="contact@fakeisthenewreal.org",

    url="https://github.com/fitnr/convertdate",

    packages=[
        "convertdate",
        "convertdate.data"
    ],

    test_suite='tests',

    zip_safe=True,

    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Religion',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License'

    ],

    install_requires=[
        'ephem>=3.7.5.3, <3.8',
        'pytz >= 2014.10, <2016'
    ]
)



# -*- coding: utf-8 -*-

# Learn more: https://github.com/kennethreitz/setup.py

import setuptools
import os

from io import open

here = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

with open(os.path.join(here, 'LICENSE.txt'), encoding='utf-8') as f:
    license_text = f.read()

setuptools.setup(
    name='RinexParser',
    version='1.0.3',
    description='Parsing Rinex files (supports version 2+3)',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Juergen Fredriksson',
    author_email='juergen.fredriksson@bev.gv.at',
    url='https://gitlab.com/dach.pos/rinexparser',
    # license=license_text,
    # packages=['rinex_parser'],
    packages=setuptools.find_packages(exclude=('tests', 'docs', 'env', '.vscode')),
    keywords='rinex parser rinex2 rinex3 gnss gps galileo beidou glonass',
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Operating System :: POSIX :: Linux",
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the GPL-v3.0 license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from math import trunc
from . import gregorian
from .utils import jwday, monthcalendarhelper

EPOCH = 347995.5
HEBREW_YEAR_OFFSET = 3760

# Hebrew months
NISAN = 1
IYYAR = 2
SIVAN = 3
TAMMUZ = 4
AV = 5
ELUL = 6
TISHRI = 7
HESHVAN = 8
KISLEV = 9
TEVETH = 10
SHEVAT = 11
ADAR = 12
VEADAR = 13


def leap(year):
    #//  Is a given Hebrew year a leap year ?
    return (((year * 7) + 1) % 19) < 7


def year_months(year):
    '''How many months are there in a Hebrew year (12 = normal, 13 = leap)'''
    if leap(year):
        return 13
    else:
        return 12


def delay_1(year):
    '''Test for delay of start of new year and to avoid'''
    #//  Sunday, Wednesday, and Friday as start of the new year.
    months = trunc(((235 * year) - 234) / 19)
    parts = 12084 + (13753 * months)
    day = trunc((months * 29) + parts / 25920)

    if ((3 * (day + 1)) % 7) < 3:
        day += 1

    return day


def delay_2(year):
    '''Check for delay in start of new year due to length of adjacent years'''

    last = delay_1(year - 1)
    present = delay_1(year)
    next_ = delay_1(year + 1)

    if next_ - present == 356:
        return 2
    elif present - last == 382:
        return 1
    else:
        return 0


def year_days(year):
    '''How many days are in a Hebrew year ?'''
    return to_jd(year + 1, 7, 1) - to_jd(year, 7, 1)


def month_days(year, month):
    '''How many days are in a given month of a given year'''
    if month > 13:
        raise ValueError("Incorrect month index")

    # First of all, dispose of fixed-length 29 day months
    if month in (IYYAR, TAMMUZ, ELUL, TEVETH, VEADAR):
        return 29

    # If it's not a leap year, Adar has 29 days
    if month == ADAR and not leap(year):
        return 29

    # If it's Heshvan, days depend on length of year
    if month == HESHVAN and (year_days(year) % 10) != 5:
        return 29

    # Similarly, Kislev varies with the length of year
    if month == KISLEV and (year_days(year) % 10) == 3:
        return 29

    # Nope, it's a 30 day month
    return 30


def to_jd(year, month, day):
    months = year_months(year)
    jd = EPOCH + delay_1(year) + delay_2(year) + day + 1

    if month < 7:
        for mon in range(7, months + 1):
            jd += month_days(year, mon)

        for mon in range(1, month):
            jd += month_days(year, mon)
    else:
        for mon in range(7, month):
            jd += month_days(year, mon)

    return int(jd) + 0.5


def from_jd(jd):
    jd = trunc(jd) + 0.5
    count = trunc(((jd - EPOCH) * 98496.0) / 35975351.0)
    year = count - 1
    i = count
    while jd >= to_jd(i, 7, 1):
        i += 1
        year += 1

    if jd < to_jd(year, 1, 1):
        first = 7
    else:
        first = 1

    month = i = first
    while jd > to_jd(year, i, month_days(year, i)):
        i += 1
        month += 1

    day = int(jd - to_jd(year, month, 1)) + 1
    return (year, month, day)


def to_jd_gregorianyear(gregorianyear, hebrew_month, hebrew_day):
    # gregorian year is either 3760 or 3761 years less than hebrew year
    # we'll first try 3760 if conversion to gregorian isn't the same
    # year that was passed to this method, then it must be 3761.

    for y in (gregorianyear + HEBREW_YEAR_OFFSET, gregorianyear + HEBREW_YEAR_OFFSET + 1):
        jd = to_jd(y, hebrew_month, hebrew_day)
        gd = gregorian.from_jd(jd)
        if gd[0] == gregorianyear:
            break
        else:
            gd = None

    if not gd:  # should never occur, but just incase...
        raise ValueError("Could not determine gregorian year")

    # tuple: (y, m, d)
    return (gd[0], gd[1], gd[2])


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_days(year, month)
    return monthcalendarhelper(start_weekday, monthlen)
